import functools
import random
import warnings
from pathlib import Path
from typing import List, Union

import numpy as np
from numpy.typing import NDArray
from scipy.signal import convolve
import itertools

from audiomentations.core.audio_loading_utils import load_sound_file
from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import find_audio_files_in_paths


class ApplyImpulseResponse(BaseWaveformTransform):
    """Convolve the audio with a randomly selected impulse response.
    Impulse responses can be created using e.g. http://tulrich.com/recording/ir_capture/
    Impulse responses are represented as audio (ideally wav) files in the given ir_path.
    """

    supports_multichannel = True

    def __init__(
        self,
        ir_path: Union[List[Path], List[str], str, Path],
        p=0.5,
        lru_cache_size=128,
        leave_length_unchanged: bool = True,
    ):
        """
        :param ir_path: A path or list of paths to audio file(s) and/or folder(s) with
            audio files. Can be str or Path instance(s). The audio files given here are
            supposed to be impulse responses.
        :param p: The probability of applying this transform
        :param lru_cache_size: Maximum size of the LRU cache for storing impulse response files
        in memory.
        :param leave_length_unchanged: When set to True, the tail of the sound (e.g. reverb at
            the end) will be chopped off so that the length of the output is equal to the
            length of the input.
        """
        super().__init__(p)
        self.ir_path = ir_path
        self.ir_files = [str(p) for p in find_audio_files_in_paths(self.ir_path)]
        assert self.ir_files, "No impulse response files found at the specified path."
        self.lru_cache_size = lru_cache_size
        self.__load_ir = functools.lru_cache(maxsize=self.lru_cache_size)(self.__load_ir)
        self.leave_length_unchanged = leave_length_unchanged

    @staticmethod
    def __load_ir(file_path, sample_rate, mono):
        return load_sound_file(file_path, sample_rate, mono=mono)

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["ir_file_path"] = random.choice(self.ir_files)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        # Determine if the impulse response should be loaded as mono
        load_mono_ir = samples.ndim == 1
        ir, sample_rate2 = self.__load_ir(self.parameters["ir_file_path"], sample_rate, mono=load_mono_ir)
        if sample_rate != sample_rate2:
            # This will typically not happen, as librosa should automatically resample the
            # impulse response sound to the desired sample rate
            raise Exception(
                "Recording sample rate {} did not match Impulse Response signal"
                " sample rate {}!".format(sample_rate, sample_rate2)
            )

        # Expand dimensions to match
        samples_original_dim = samples.ndim
        samples, ir = np.atleast_2d(samples), np.atleast_2d(ir)

        # Preallocate the output array
        output_shape = (samples.shape[0], samples.shape[1] + ir.shape[1] - 1)
        signal_ir = np.empty(output_shape, dtype=samples.dtype)

        # Loop over all samples channels for channelwise convolution
        for i, (sample, impulse_response) in enumerate(zip(samples, itertools.cycle(ir))):
            signal_ir[i, :] = convolve(sample, impulse_response)

        max_value = max(np.amax(signal_ir), -np.amin(signal_ir))
        if max_value > 0.0:
            scale = 0.5 / max_value
            signal_ir *= scale
        if self.leave_length_unchanged:
            signal_ir = signal_ir[..., : samples.shape[-1]]

        # reshape if mono input
        if samples_original_dim == 1:
            signal_ir = signal_ir[0]

        return signal_ir

    def __getstate__(self):
        state = self.__dict__.copy()
        warnings.warn(
            "Warning: the LRU cache of ApplyImpulseResponse gets discarded when pickling it."
            " E.g. this means the cache will be not be used when using ApplyImpulseResponse"
            " together with multiprocessing on Windows"
        )
        del state["_ApplyImpulseResponse__load_ir"]
        return state



import numpy as np
import pytest

from audiomentations import Lambda, Gain


class TestLambda:
    def test_gain_lambda(self):
        samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
        augmenter = Lambda(
            transform=Gain(min_gain_db=50, max_gain_db=50, p=1.0), p=1.0
        )
        std_in = np.mean(np.abs(samples_in))
        samples_out = augmenter(samples=samples_in, sample_rate=16000)
        std_out = np.mean(np.abs(samples_out))
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape
        assert std_out > 100 * std_in

    def test_lambda_with_kwargs(self):
        samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
        augmenter = Lambda(
            transform=lambda samples, sample_rate, offset: samples + offset,
            p=1.0,
            offset=-0.2,
        )
        input_mean = np.mean(samples_in)
        samples_out = augmenter(samples=samples_in, sample_rate=16000)
        output_mean = np.mean(samples_out)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape
        assert output_mean == pytest.approx(input_mean - 0.2)



import os
import pickle

import numpy as np

from audiomentations import ApplyImpulseResponse
from audiomentations.core.composition import Compose
from demo.demo import DEMO_DIR


class TestImpulseResponse:
    def test_apply_impulse_response(self):
        sample_len = 1024
        samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000

        add_ir_transform = ApplyImpulseResponse(
            ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0
        )

        # Check that misc_file.txt is not one of the IR file candidates, as it's not audio
        assert len(add_ir_transform.ir_files) == 3

        augmenter = Compose([add_ir_transform])

        assert len(samples_in) == sample_len
        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)

        # Check parameters
        assert augmenter.transforms[0].parameters["should_apply"]
        assert augmenter.transforms[0].parameters["ir_file_path"].endswith(".wav")

        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape

    def test_apply_impulse_response_multi_channel(self):
        sample_len = 1024
        samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000

        samples_in = np.expand_dims(samples_in, axis=0)
        samples_in = np.tile(samples_in, (2, 1))

        add_ir_transform = ApplyImpulseResponse(
            ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0
        )

        # Check that misc_file.txt is not one of the IR file candidates, as it's not audio
        assert len(add_ir_transform.ir_files) == 3

        augmenter = Compose([add_ir_transform])

        assert samples_in.shape[1] == sample_len
        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)

        # Check parameters
        assert augmenter.transforms[0].parameters["should_apply"]
        assert augmenter.transforms[0].parameters["ir_file_path"].endswith(".wav")

        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape

    def test_include_tail(self):
        sample_len = 1024
        samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000

        add_ir_transform = ApplyImpulseResponse(
            ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0, leave_length_unchanged=False
        )

        samples_out = add_ir_transform(samples=samples_in, sample_rate=sample_rate)

        assert samples_out.dtype == np.float32
        assert samples_out.shape[-1] > samples_in.shape[-1]

    def test_picklability(self):
        add_ir_transform = ApplyImpulseResponse(
            ir_path=os.path.join(DEMO_DIR, "ir"), p=1.0
        )
        pickled = pickle.dumps(add_ir_transform)
        unpickled = pickle.loads(pickled)
        assert add_ir_transform.ir_files == unpickled.ir_files



import random

import numpy as np
import pytest

from audiomentations import AdjustDuration


class TestAdjustDuration:
    @pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
    @pytest.mark.parametrize("pad_section", ["start", "end"])
    @pytest.mark.parametrize("sample_len", [3, 4, 5])
    @pytest.mark.parametrize("ndim", [None, 1, 2])
    def test_padding(self, mode, pad_section, sample_len, ndim):
        random.seed(546)
        samples = np.ones((ndim, 4) if ndim else 4, dtype=np.float32)
        sample_rate = 16000
        input_shape = samples.shape
        target_shape = list(input_shape)
        target_shape[-1] = sample_len
        target_shape = tuple(target_shape)
        augmenter = AdjustDuration(
            duration_samples=sample_len, padding_mode=mode, padding_position=pad_section, p=1.0
        )
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape == target_shape

    @pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
    @pytest.mark.parametrize("pad_section", ["start", "end"])
    @pytest.mark.parametrize("second", [0.4, 0.5, 0.6])
    @pytest.mark.parametrize("ndim", [None, 1, 2])
    def test_padding_second(self, mode, pad_section, second, ndim):
        random.seed(546)
        sample_rate = 80
        samples = np.ones((ndim, 40) if ndim else 40, dtype=np.float32)
        input_shape = samples.shape
        target_shape = list(input_shape)
        target_shape[-1] = int(second * sample_rate)
        target_shape = tuple(target_shape)
        augmenter = AdjustDuration(
            duration_seconds=second, padding_mode=mode, padding_position=pad_section, p=1.0
        )
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape == target_shape



import random
from typing import Union

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import convert_decibels_to_amplitude_ratio


def get_fade_mask(
    start_level_db: float,
    end_level_db: float,
    fade_time_samples: int,
):
    """
    :param start_level_db:
    :param end_level_db:
    :param fade_time_samples: How long does the fade last?
    :return:
    """
    fade_mask = np.linspace(
        start_level_db,
        end_level_db,
        num=fade_time_samples,
        dtype=np.float32,
    )
    fade_mask = convert_decibels_to_amplitude_ratio(fade_mask)
    return fade_mask


class GainTransition(BaseWaveformTransform):
    """
    Gradually change the volume up or down over a random time span. Also known as
    fade in and fade out. The fade works on a logarithmic scale, which is natural to
    human hearing.

    The way this works is that it picks two gains: a first gain and a second gain.
    Then it picks a time range for the transition between those two gains.
    Note that this transition can start before the audio starts and/or end after the
    audio ends, so the output audio can start or end in the middle of a transition.
    The gain starts at the first gain and is held constant until the transition start.
    Then it transitions to the second gain. Then that gain is held constant until the
    end of the sound.
    """

    supports_multichannel = True

    def __init__(
        self,
        min_gain_db: float = -24.0,
        max_gain_db: float = 6.0,
        min_duration: Union[float, int] = 0.2,
        max_duration: Union[float, int] = 6.0,
        duration_unit: str = "seconds",
        p: float = 0.5,
    ):
        """
        :param min_gain_db: Minimum gain in dB.
        :param max_gain_db: Maximum gain in dB.
        :param min_duration: Minimum length of transition. See also duration_unit.
        :param max_duration: Maximum length of transition. See also duration_unit.
        :param duration_unit: Defines the unit of the value of min_duration and max_duration.
            "fraction": Fraction of the total sound length
            "samples": Number of audio samples
            "seconds": Number of seconds
        :param p: The probability of applying this transform
        """
        super().__init__(p)

        if min_gain_db > max_gain_db:
            raise ValueError("min_gain_db must not be greater than max_gain_db")
        self.min_gain_db = min_gain_db
        self.max_gain_db = max_gain_db

        if min_duration <= 0:
            raise ValueError("min_duration must be greater than zero")
        if min_duration > max_duration:
            raise ValueError("min_duration must not be greater than max_duration")
        self.min_duration = min_duration
        self.max_duration = max_duration
        self.duration_unit = duration_unit

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            if self.duration_unit == "samples":
                min_duration_in_samples = self.min_duration
                max_duration_in_samples = self.max_duration
            elif self.duration_unit == "fraction":
                min_duration_in_samples = int(
                    round(self.min_duration * samples.shape[-1])
                )
                max_duration_in_samples = int(
                    round(self.max_duration * samples.shape[-1])
                )
            elif self.duration_unit == "seconds":
                min_duration_in_samples = int(round(self.min_duration * sample_rate))
                max_duration_in_samples = int(round(self.max_duration * sample_rate))
            else:
                raise ValueError("Invalid duration_unit")

            self.parameters["fade_time_samples"] = max(
                3, random.randint(min_duration_in_samples, max_duration_in_samples)
            )
            self.parameters["t0"] = random.randint(
                -self.parameters["fade_time_samples"] + 2,
                samples.shape[-1] - 2,
                )
            self.parameters["start_gain_db"] = random.uniform(
                self.min_gain_db, self.max_gain_db
            )
            self.parameters["end_gain_db"] = random.uniform(
                self.min_gain_db, self.max_gain_db
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        num_samples = samples.shape[-1]
        fade_mask = get_fade_mask(
            start_level_db=self.parameters["start_gain_db"],
            end_level_db=self.parameters["end_gain_db"],
            fade_time_samples=self.parameters["fade_time_samples"],
        )
        start_sample_index = self.parameters["t0"]
        end_sample_index = start_sample_index + self.parameters["fade_time_samples"]
        if start_sample_index < 0:
            # crop fade_mask: shave off a chunk in the beginning
            num_samples_to_shave_off = abs(start_sample_index)
            fade_mask = fade_mask[num_samples_to_shave_off:]
            start_sample_index = 0

        if end_sample_index > num_samples:
            # crop fade_mask: shave off a chunk in the end
            num_samples_to_shave_off = end_sample_index - num_samples
            fade_mask = fade_mask[: fade_mask.shape[-1] - num_samples_to_shave_off]
            end_sample_index = num_samples

        samples = np.copy(samples)

        samples[..., start_sample_index:end_sample_index] *= fade_mask
        if start_sample_index > 0:
            samples[..., :start_sample_index] *= convert_decibels_to_amplitude_ratio(
                self.parameters["start_gain_db"]
            )
        if end_sample_index < num_samples:
            samples[..., end_sample_index:] *= convert_decibels_to_amplitude_ratio(
                self.parameters["end_gain_db"]
            )
        return samples



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class BitCrush(BaseWaveformTransform):
    """
    Apply a bit crush effect to the audio by reducing the bit depth. In other words, it
    reduces the number of bits that can be used for representing each audio sample.
    This adds quantization noise, and affects dynamic range. This transform does not
    apply dithering.
    """

    supports_multichannel = True

    def __init__(self, min_bit_depth: int = 5, max_bit_depth: int = 10, p: float = 0.5):
        """
        :param min_bit_depth: The minimum bit depth the audio will be "converted" to
        :param max_bit_depth: The maximum bit depth the audio will be "converted" to
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        self.min_bit_depth = min_bit_depth
        self.max_bit_depth = max_bit_depth

        if min_bit_depth < 1:
            raise ValueError("min_bit_depth must be at least 1")

        if max_bit_depth > 32:
            raise ValueError("max_bit_depth must not be greater than 32")

        if min_bit_depth > max_bit_depth:
            raise ValueError("min_bit_depth must not be larger than max_bit_depth")

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["bit_depth"] = random.randint(
                self.min_bit_depth, self.max_bit_depth
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        q = (2 ** self.parameters["bit_depth"] / 2) + 1
        return np.round(samples * q) / q



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class Padding(BaseWaveformTransform):
    """
    Apply padding to the audio signal by taking a fraction of the start or end and replacing that
    portion with padding. This can be useful for training ML models on padded inputs.
    """

    supports_multichannel = True

    def __init__(
        self,
        mode: str = "silence",
        min_fraction: float = 0.01,
        max_fraction: float = 0.7,
        pad_section: str = "end",
        p: float = 0.5,
    ):
        """
        :param mode: Padding mode. Must be one of "silence", "wrap", "reflect"
        :param min_fraction: Minimum fraction of the signal duration to be padded
        :param max_fraction: Maximum fraction of the signal duration to be padded
        :param pad_section: Which part of the signal should be replaced with padding:
            "start" or "end"
        :param p: The probability of applying this transform
        """
        super().__init__(p)

        assert mode in ("silence", "wrap", "reflect")
        self.mode = mode

        assert max_fraction <= 1.0
        assert min_fraction >= 0
        assert min_fraction <= max_fraction
        self.min_fraction = min_fraction
        self.max_fraction = max_fraction

        assert pad_section in ("start", "end")
        self.pad_section = pad_section

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            input_length = samples.shape[-1]
            self.parameters["padding_length"] = random.randint(
                int(round(self.min_fraction * input_length)),
                int(round(self.max_fraction * input_length)),
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        padding_length = self.parameters["padding_length"]
        if padding_length == 0:
            return samples

        untouched_length = samples.shape[-1] - padding_length

        if self.mode == "silence":
            samples = np.copy(samples)
            if self.pad_section == "start":
                samples[..., :padding_length] = 0.0
            else:
                samples[..., -padding_length:] = 0.0
        else:
            if samples.ndim == 1:
                if self.pad_section == "start":
                    pad_width = (padding_length, 0)
                else:
                    pad_width = (0, padding_length)
            else:
                if self.pad_section == "start":
                    pad_width = ((0, 0), (padding_length, 0))
                else:
                    pad_width = ((0, 0), (0, padding_length))

            if self.pad_section == "start":
                samples = samples[..., -untouched_length:]
            else:
                samples = samples[..., :untouched_length]

            samples = np.pad(samples, pad_width, self.mode)

        return samples



import numpy as np
from numpy.testing import assert_array_almost_equal

from audiomentations import Reverse


class TestReverse:
    def test_single_channel(self):
        samples = np.array([0.5, 0.6, -0.2, 0.0], dtype=np.float32)
        sample_rate = 16000
        augmenter = Reverse(p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape[-1] == 4

    def test_multichannel(self):
        samples = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        sample_rate = 16000
        augmenter = Reverse(p=1.0)
        reversed_samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert reversed_samples.dtype == np.float32
        assert_array_almost_equal(
            reversed_samples,
            np.array(
                [[0.0, -0.125, -0.25, 0.5, 0.9], [0.0, -0.125, -0.25, 0.5, 0.95]],
                dtype=np.float32,
            ),
        )



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class ClippingDistortion(BaseWaveformTransform):
    """Distort signal by clipping a random percentage of points

    The percentage of points that will be clipped is drawn from a uniform distribution between
    the two input parameters min_percentile_threshold and max_percentile_threshold. If for instance
    30% is drawn, the samples are clipped if they're below the 15th or above the 85th percentile.
    """

    supports_multichannel = True

    def __init__(
        self,
        min_percentile_threshold: int = 0,
        max_percentile_threshold: int = 40,
        p: float = 0.5,
    ):
        """
        :param min_percentile_threshold: int, A lower bound on the total percent of samples that
            will be clipped
        :param max_percentile_threshold: int, An upper bound on the total percent of samples that
            will be clipped
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert min_percentile_threshold <= max_percentile_threshold
        assert 0 <= min_percentile_threshold <= 100
        assert 0 <= max_percentile_threshold <= 100
        self.min_percentile_threshold = min_percentile_threshold
        self.max_percentile_threshold = max_percentile_threshold

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["percentile_threshold"] = random.randint(
                self.min_percentile_threshold, self.max_percentile_threshold
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        lower_percentile_threshold = int(self.parameters["percentile_threshold"] / 2)
        lower_threshold, upper_threshold = np.percentile(
            samples, [lower_percentile_threshold, 100 - lower_percentile_threshold]
        )
        samples = np.clip(samples, lower_threshold, upper_threshold)
        return samples



from audiomentations import Normalize


class TestTransformsInterface:
    def test_freeze_and_unfreeze_parameters(self):
        normalizer = Normalize(p=1.0)

        assert normalizer.are_parameters_frozen == False

        normalizer.freeze_parameters()
        assert normalizer.are_parameters_frozen == True

        normalizer.unfreeze_parameters()
        assert normalizer.are_parameters_frozen == False



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from audiomentations import Shift


class TestShift:
    def test_shift_fraction(self):
        samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
        sample_rate = 16000

        forward_augmenter = Shift(
            min_shift=0.5, max_shift=0.5, fade_duration=0.0, p=1.0
        )
        forward_shifted_samples = forward_augmenter(
            samples=samples, sample_rate=sample_rate
        )
        assert_almost_equal(
            forward_shifted_samples, np.array([0.25, 0.125, 1.0, 0.5], dtype=np.float32)
        )
        assert forward_shifted_samples.dtype == np.float32
        assert len(forward_shifted_samples) == 4

        backward_augmenter = Shift(
            min_shift=-0.25, max_shift=-0.25, fade_duration=0.0, p=1.0
        )
        backward_shifted_samples = backward_augmenter(
            samples=samples, sample_rate=sample_rate
        )
        assert_almost_equal(
            backward_shifted_samples,
            np.array([0.5, 0.25, 0.125, 1.0], dtype=np.float32),
        )
        assert backward_shifted_samples.dtype == np.float32
        assert len(forward_shifted_samples) == 4

    def test_shift_samples(self):
        samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
        sample_rate = 16000

        forward_augmenter = Shift(
            min_shift=1, max_shift=1, shift_unit="samples", fade_duration=0.0, p=1.0
        )
        forward_shifted_samples = forward_augmenter(
            samples=samples, sample_rate=sample_rate
        )
        assert_almost_equal(
            forward_shifted_samples, np.array([0.125, 1.0, 0.5, 0.25], dtype=np.float32)
        )
        assert forward_shifted_samples.dtype == np.float32
        assert len(forward_shifted_samples) == 4

    def test_shift_seconds(self):
        samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
        sample_rate = 2

        forward_augmenter = Shift(
            min_shift=1.0, max_shift=1.0, shift_unit="seconds", fade_duration=0.0, p=1.0
        )
        forward_shifted_samples = forward_augmenter(
            samples=samples, sample_rate=sample_rate
        )
        assert_almost_equal(
            forward_shifted_samples, np.array([0.25, 0.125, 1.0, 0.5], dtype=np.float32)
        )
        assert forward_shifted_samples.dtype == np.float32
        assert len(forward_shifted_samples) == 4

    def test_shift_without_rollover(self):
        samples = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
        sample_rate = 16000

        forward_augmenter = Shift(
            min_shift=0.5, max_shift=0.5, rollover=False, fade_duration=0.0, p=1.0
        )
        forward_shifted_samples = forward_augmenter(
            samples=samples, sample_rate=sample_rate
        )
        assert_almost_equal(
            forward_shifted_samples, np.array([0.0, 0.0, 1.0, 0.5], dtype=np.float32)
        )
        assert forward_shifted_samples.dtype == np.float32
        assert len(forward_shifted_samples) == 4

        backward_augmenter = Shift(
            min_shift=-0.25, max_shift=-0.25, rollover=False, fade_duration=0.0, p=1.0
        )
        backward_shifted_samples = backward_augmenter(
            samples=samples, sample_rate=sample_rate
        )
        assert_almost_equal(
            backward_shifted_samples,
            np.array([0.5, 0.25, 0.125, 0.0], dtype=np.float32),
        )
        assert backward_shifted_samples.dtype == np.float32
        assert len(forward_shifted_samples) == 4

    def test_shift_multichannel(self):
        samples = np.array(
            [[0.75, 0.5, -0.25, -0.125], [0.9, 0.5, -0.25, -0.125]], dtype=np.float32
        )
        sample_rate = 4000

        augment = Shift(min_shift=0.5, max_shift=0.5, fade_duration=0.0, p=1.0)
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        assert_almost_equal(
            processed_samples,
            np.array(
                [[-0.25, -0.125, 0.75, 0.5], [-0.25, -0.125, 0.9, 0.5]],
                dtype=np.float32,
            ),
        )
        assert processed_samples.dtype == np.float32

    def test_shift_without_rollover_multichannel(self):
        samples = np.array(
            [[0.75, 0.5, -0.25, -0.125], [0.9, 0.5, -0.25, -0.125]], dtype=np.float32
        )
        sample_rate = 4000

        augment = Shift(
            min_shift=0.5, max_shift=0.5, rollover=False, fade_duration=0.0, p=1.0
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        assert_almost_equal(
            processed_samples,
            np.array([[0.0, 0.0, 0.75, 0.5], [0.0, 0.0, 0.9, 0.5]], dtype=np.float32),
        )
        assert processed_samples.dtype == np.float32

    def test_shift_fade(self):
        samples = np.array(
            [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
            dtype=np.float32,
        )
        sample_rate = 4000

        augment = Shift(
            min_shift=0.5,
            max_shift=0.5,
            rollover=False,
            fade_duration=0.00075,  # 0.00075 * 4000 = 3
            p=1.0,
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        assert processed_samples == pytest.approx(
            np.array(
                [[0.0, 0.0, 0.0, 1.4067, 3.0], [0.0, 0.0, 0.0, -1.4067, -3.0]],
                dtype=np.float32,
            ),
            abs=0.01,
        )

    def test_shift_fade_rollover(self):
        samples = np.array(
            [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
            dtype=np.float32,
        )
        sample_rate = 4000

        augment = Shift(
            min_shift=0.5,
            max_shift=0.5,
            rollover=True,
            fade_duration=0.00075,  # 0.00075 * 4000 = 3
            p=1.0,
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert processed_samples == pytest.approx(
            np.array(
                [[2.81, 0.0, 0, 1.407, 3.0], [-2.81, 0.0, 0, -1.407, -3.0]],
                dtype=np.float32,
            ),
            abs=0.01,
        )

    def test_shift_fade_rollover_2(self):
        samples = np.array(
            [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
            dtype=np.float32,
        )
        sample_rate = 4000

        augment = Shift(
            min_shift=-0.5,
            max_shift=-0.5,
            rollover=True,
            fade_duration=0.00075,  # 0.00075 * 4000 = 3
            p=1.0,
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert processed_samples == pytest.approx(
            np.array(
                [[3.0, 2.81, 0.0, 0.0, 1.407], [-3.0, -2.81, 0.0, -0.0, -1.407]],
                dtype=np.float32,
            ),
            abs=0.01,
        )

    def test_shift_fade_rollover_3(self):
        samples = np.array(
            [[1.0, 2.0, 3.0, 4.0, 5.0], [-1.0, -2.0, -3.0, -4.0, -5.0]],
            dtype=np.float32,
        )
        sample_rate = 4000

        augment = Shift(
            min_shift=-0.5,
            max_shift=-0.5,
            rollover=True,
            fade_duration=1.0,
            p=1.0,
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert processed_samples == pytest.approx(
            np.array(
                [
                    [
                        3.0023373e-06,
                        1.0004364e-06,
                        0.0000000e00,
                        0.0000000e00,
                        5.0030241e-07,
                    ],
                    [
                        -3.0023373e-06,
                        -1.0004364e-06,
                        -0.0000000e00,
                        -0.0000000e00,
                        -5.0030241e-07,
                    ],
                ],
                dtype=np.float32,
            ),
        )

    def test_freeze_shift_with_different_sample_rates(self):
        samples1 = np.array([1.0, 0.5, 0.25, 0.125], dtype=np.float32)
        samples2 = np.array(
            [1.0, 0.5, 0.25, 0.125, 0.3, 0.4, 0.5, 0.2], dtype=np.float32
        )
        sample_rate_1 = 1
        sample_rate_2 = 2

        augment = Shift(
            min_shift=0.5, max_shift=0.5, rollover=False, fade_duration=0.0, p=1.0
        )
        augment.randomize_parameters(samples1, sample_rate_1)
        augment.freeze_parameters()
        shifted_samples1 = augment(samples1, sample_rate_1)
        shifted_samples2 = augment(samples2, sample_rate_2)

        num_samples1_shifted = np.count_nonzero(shifted_samples1 == 0)
        num_samples2_shifted = np.count_nonzero(shifted_samples2 == 0)

        assert num_samples1_shifted == 2
        assert num_samples2_shifted == 4

    def test_invalid_parameters(self):
        with pytest.raises(ValueError):
            Shift(min_shift=-1.5)
        with pytest.raises(ValueError):
            Shift(min_shift=1.5)
        with pytest.raises(ValueError):
            Shift(max_shift=-1.5)
        with pytest.raises(ValueError):
            Shift(max_shift=1.5)
        with pytest.raises(ValueError):
            Shift(min_shift=0.2, max_shift=0.1)
        with pytest.raises(ValueError):
            Shift(fade_duration=0.000001)
        with pytest.raises(ValueError):
            Shift(fade_duration=-1337)
        with pytest.raises(ValueError):
            Shift(shift_unit="lightyears")



import math
import random
import sys

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import (
    convert_decibels_to_amplitude_ratio,
    get_max_abs_amplitude,
)


class Limiter(BaseWaveformTransform):
    """
    A simple audio limiter (dynamic range compression).
    """

    supports_multichannel = True

    def __init__(
        self,
        min_threshold_db: float = -24.0,
        max_threshold_db: float = -2.0,
        min_attack: float = 0.0005,
        max_attack: float = 0.025,
        min_release: float = 0.05,
        max_release: float = 0.7,
        threshold_mode: str = "relative_to_signal_peak",
        p: float = 0.5,
    ):
        """
        The threshold determines the audio level above which the limiter kicks in.
        The attack time is how quickly the limiter kicks in once the audio signal starts exceeding the threshold.
        The release time determines how quickly the limiter stops working after the signal drops below the threshold.

        :param min_threshold_db: Minimum threshold in Decibels
        :param max_threshold_db: Maximum threshold in Decibels
        :param min_attack: Minimum attack time in seconds
        :param max_attack: Maximum attack time in seconds
        :param min_release: Minimum release time in seconds
        :param max_release: Maximum release time in seconds
        :param threshold_mode: "relative_to_signal_peak" or "absolute"
            "relative_to_signal_peak" means the threshold is relative to peak of the signal.
            "absolute" means the threshold is relative to 0 dBFS, so it doesn't depend
            on the peak of the signal.
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert min_attack > 0.0, "min_attack must be a positive number"
        assert max_attack > 0.0, "max_attack must be a positive number"
        assert (
            max_attack >= min_attack
        ), "max_attack must be greater than or equal to min_attack"
        assert min_release > 0.0, "min_release must be a positive number"
        assert max_release > 0.0, "max_release must be a positive number"
        assert (
            max_release >= min_release
        ), "max_release must be greater than or equal to min_release"
        assert (
            max_threshold_db >= min_threshold_db
        ), "max_threshold_db must be greater than or equal to min_threshold_db"
        assert threshold_mode in (
            "relative_to_signal_peak",
            "absolute",
        ), 'threshold_mode must be either "relative_to_signal_peak" or "absolute"'
        self.min_threshold_db = min_threshold_db
        self.max_threshold_db = max_threshold_db
        self.min_attack = min_attack
        self.max_attack = max_attack
        self.min_release = min_release
        self.max_release = max_release
        self.threshold_mode = threshold_mode

    @staticmethod
    def convert_time_to_coefficient(
        t: float, sample_rate: int, decay_threshold: float = None
    ) -> float:
        if decay_threshold is None:
            # Attack time and release time in this transform are defined as how long
            # it takes to step 1-decay_threshold of the way to a constant target gain.
            # The default threshold used here is inspired by RT60.
            decay_threshold = convert_decibels_to_amplitude_ratio(-60)
        return 10 ** (math.log10(decay_threshold) / max(sample_rate * t, 1.0))

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)

        if self.parameters["should_apply"]:
            attack_seconds = random.uniform(self.min_attack, self.max_attack)
            self.parameters["attack"] = self.convert_time_to_coefficient(
                attack_seconds, sample_rate
            )
            release_seconds = random.uniform(self.min_release, self.max_release)
            self.parameters["release"] = self.convert_time_to_coefficient(
                release_seconds, sample_rate
            )
            # Delay the signal by 60% of the attack time by default
            self.parameters["delay"] = max(round(0.6 * attack_seconds * sample_rate), 1)

            threshold_factor = (
                get_max_abs_amplitude(samples)
                if self.threshold_mode == "relative_to_signal_peak"
                else 1.0
            )
            threshold_db = random.uniform(self.min_threshold_db, self.max_threshold_db)

            self.parameters[
                "threshold"
            ] = threshold_factor * convert_decibels_to_amplitude_ratio(threshold_db)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        if self.parameters["threshold"] == 0.0:
            # Digital silence input can cause this to happen
            return samples
        try:
            from cylimiter import Limiter as CyLimiter
        except ImportError:
            print(
                "Failed to import cylimiter. Maybe it is not installed? "
                "To install the optional cylimiter dependency of audiomentations,"
                " run `pip install cylimiter` or `pip install audiomentations[extras]`",
                file=sys.stderr,
            )
            raise

        limiter = CyLimiter(
            attack=self.parameters["attack"],
            release=self.parameters["release"],
            delay=self.parameters["delay"],
            threshold=self.parameters["threshold"],
        )

        if samples.ndim == 1:
            processed_samples = np.pad(samples, (0, self.parameters["delay"]))
            limiter.limit_inplace(processed_samples)
            processed_samples = processed_samples[self.parameters["delay"] - 1 : -1]
        else:
            # By default, there is no interchannel linking. The channels are processed
            # independently. Support for linking may be added in the future:
            # https://github.com/pzelasko/cylimiter/issues/4
            processed_samples = np.copy(samples)
            for chn_idx in range(samples.shape[0]):
                limiter.reset()
                channel = np.ascontiguousarray(
                    np.pad(processed_samples[chn_idx, :], (0, self.parameters["delay"]))
                )
                limiter.limit_inplace(channel)
                processed_samples[chn_idx, :] = channel[
                    self.parameters["delay"] - 1 : -1
                ]

        return processed_samples



import random

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class TimeMask(BaseWaveformTransform):
    """
    Make a randomly chosen part of the audio silent.
    Inspired by https://arxiv.org/pdf/1904.08779.pdf
    """

    supports_multichannel = True

    def __init__(
        self,
        min_band_part: float = 0.0,
        max_band_part: float = 0.5,
        fade: bool = False,
        p: float = 0.5,
    ):
        """
        :param min_band_part: Minimum length of the silent part as a fraction of the
            total sound length. Must be between 0.0 and 1.0
        :param max_band_part: Maximum length of the silent part as a fraction of the
            total sound length. Must be between 0.0 and 1.0
        :param fade: When set to True, a linear fade-in and fade-out is added to the silent part.
            This can smooth out unwanted abrupt changes between consecutive samples, which might
            otherwise sound like transients/clicks/pops.
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        if min_band_part < 0.0 or min_band_part > 1.0:
            raise ValueError("min_band_part must be between 0.0 and 1.0")
        if max_band_part < 0.0 or max_band_part > 1.0:
            raise ValueError("max_band_part must be between 0.0 and 1.0")
        if min_band_part > max_band_part:
            raise ValueError("min_band_part must not be greater than max_band_part")
        self.min_band_part = min_band_part
        self.max_band_part = max_band_part
        self.fade = fade

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            num_samples = samples.shape[-1]
            self.parameters["t"] = random.randint(
                int(num_samples * self.min_band_part),
                int(num_samples * self.max_band_part),
            )
            self.parameters["t0"] = random.randint(
                0, num_samples - self.parameters["t"]
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        new_samples = samples.copy()
        t = self.parameters["t"]
        t0 = self.parameters["t0"]
        mask = np.zeros(t)
        if self.fade:
            fade_length = min(int(sample_rate * 0.01), int(t * 0.1))
            mask[0:fade_length] = np.linspace(1, 0, num=fade_length)
            mask[-fade_length:] = np.linspace(0, 1, num=fade_length)
        new_samples[..., t0 : t0 + t] *= mask
        return new_samples



import math
import os
from functools import lru_cache
from pathlib import Path
from typing import List, Union, Tuple, Any

import numpy as np
import numpy_minmax
import numpy_rms
from numpy.typing import NDArray

SUPPORTED_EXTENSIONS = (
    ".aac",
    ".aif",
    ".aiff",
    ".flac",
    ".m4a",
    ".mp3",
    ".mp4",
    ".ogg",
    ".opus",
    ".wav",
)

def format_args(args_dict: dict[str, Any]) -> str:
    formatted_args = []
    for k, v in args_dict.items():
        v_formatted = f"'{v}'" if isinstance(v, str) else str(v)
        formatted_args.append(f"{k}={v_formatted}")
    return ", ".join(formatted_args)

def find_audio_files(
    root_path,
    filename_endings=SUPPORTED_EXTENSIONS,
    traverse_subdirectories=True,
    follow_symlinks=True,
):
    """Return a list of paths to all audio files with the given extension(s) in a directory.
    Also traverses subdirectories by default.
    """
    file_paths = []

    for root, dirs, filenames in os.walk(root_path, followlinks=follow_symlinks):
        filenames = sorted(filenames)
        for filename in filenames:
            input_path = os.path.abspath(root)
            file_path = os.path.join(input_path, filename)

            if filename.lower().endswith(filename_endings):
                file_paths.append(Path(file_path))
        if not traverse_subdirectories:
            # prevent descending into subfolders
            break

    return file_paths


def find_audio_files_in_paths(
    paths: Union[List[Path], List[str], Path, str],
    filename_endings=SUPPORTED_EXTENSIONS,
    traverse_subdirectories=True,
    follow_symlinks=True,
):
    """Return a list of paths to all audio files with the given extension(s) contained in the list or in its directories.
    Also traverses subdirectories by default.
    """

    file_paths = []

    if isinstance(paths, (list, tuple, set)):
        paths = list(paths)
    else:
        paths = [paths]

    for p in paths:
        if str(p).lower().endswith(SUPPORTED_EXTENSIONS):
            file_path = Path(os.path.abspath(p))
            file_paths.append(file_path)
        elif os.path.isdir(p):
            file_paths += find_audio_files(
                p,
                filename_endings=filename_endings,
                traverse_subdirectories=traverse_subdirectories,
                follow_symlinks=follow_symlinks,
            )
    return file_paths


def calculate_rms(samples):
    """Given a numpy array of audio samples, return its Root Mean Square (RMS)."""
    return np.mean(numpy_rms.rms(samples))


def calculate_rms_without_silence(samples, sample_rate):
    """
    This function returns the rms of a given noise whose silent periods have been removed. This ensures
    that the rms of the noise is not underestimated. Is most useful for short non-stationary noises.
    """

    window = int(0.025 * sample_rate)

    if samples.shape[-1] < window:
        return calculate_rms(samples)

    rms_all_windows = np.zeros(samples.shape[-1] // window)
    current_time = 0

    while current_time < samples.shape[-1] - window:
        rms_all_windows[current_time // window] += calculate_rms(
            samples[current_time : current_time + window]
        )
        current_time += window

    rms_threshold = np.max(rms_all_windows) / 25

    # The segments with a too low rms are identified and discarded
    rms_all_windows = rms_all_windows[rms_all_windows > rms_threshold]
    # Beware that each window must have the same number of samples so that this calculation of the rms is valid.
    return calculate_rms(rms_all_windows)


def calculate_desired_noise_rms(clean_rms, snr):
    """
    Given the Root Mean Square (RMS) of a clean sound and a desired signal-to-noise ratio (SNR),
    calculate the desired RMS of a noise sound to be mixed in.
    Based on https://github.com/Sato-Kunihiko/audio-SNR/blob/8d2c933b6c0afe6f1203251f4877e7a1068a6130/create_mixed_audio_file.py#L20
    :param clean_rms: Root Mean Square (RMS) - a value between 0.0 and 1.0
    :param snr: Signal-to-Noise (SNR) Ratio in dB - typically somewhere between -20 and 60
    :return:
    """
    a = float(snr) / 20
    noise_rms = clean_rms / (10**a)
    return noise_rms


def convert_decibels_to_amplitude_ratio(decibels):
    return 10 ** (decibels / 20)


def is_waveform_multichannel(samples):
    """
    Return bool that answers the question: Is the given ndarray a multichannel waveform or not?
    :param samples: numpy ndarray
    :return:
    """
    return len(samples.shape) > 1


def is_spectrogram_multichannel(spectrogram):
    """
    Return bool that answers the question: Is the given ndarray a multichannel spectrogram?
    :param samples: numpy ndarray
    :return:
    """
    return len(spectrogram.shape) > 2 and spectrogram.shape[-1] > 1


def convert_float_samples_to_int16(y):
    """Convert floating-point numpy array of audio samples to int16."""
    if not issubclass(y.dtype.type, np.floating):
        raise ValueError("input samples not floating-point")
    return (y * np.iinfo(np.int16).max).astype(np.int16)


def convert_frequency_to_mel(f: float) -> float:
    """
    Convert f hertz to mels
    https://en.wikipedia.org/wiki/Mel_scale#Formula
    """
    return 2595.0 * math.log10(1.0 + f / 700.0)


def convert_mel_to_frequency(m: Union[float, np.array]) -> Union[float, np.array]:
    """
    Convert m mels to hertz
    https://en.wikipedia.org/wiki/Mel_scale#History_and_other_formulas
    """
    return 700.0 * (10 ** (m / 2595.0) - 1.0)


def a_weighting_frequency_envelope(n_fft, sample_rate):
    """
    Return the A-weighting frequency envelope for the given FFT size and sample rate.

    See the wikipedia article here:
    https://en.wikipedia.org/wiki/A-weighting#A
    """

    freqs = np.fft.rfftfreq(n_fft, 1 / sample_rate)
    weighting = (
        (12194**2 * freqs**4)
        / (
            (freqs**2 + 20.6**2)
            * np.sqrt((freqs**2 + 107.7**2) * (freqs**2 + 737.9**2))
            * (freqs**2 + 12194**2)
        )
    ) + 2.00
    return weighting


@lru_cache(maxsize=8)
def get_crossfade_mask_pair(
    length: int, equal_energy=True
) -> Tuple[np.array, np.array]:
    """
    Equal-gain or equal-energy (within ~1%) cross-fade mask pair with
    smooth start and end.
    https://signalsmith-audio.co.uk/writing/2021/cheap-energy-crossfade/
    """
    x = np.linspace(0, 1, length, dtype=np.float32)
    x2 = 1 - x
    a = x * x2
    k = 1.4186 if equal_energy else -0.70912
    b = a * (1 + k * a)
    c = b + x
    d = b + x2
    fade_in = c * c
    fade_out = d * d
    return fade_in, fade_out


def get_max_abs_amplitude(samples: NDArray):
    min_amplitude, max_amplitude = numpy_minmax.minmax(samples)
    max_abs_amplitude = max(abs(min_amplitude), abs(max_amplitude))
    return max_abs_amplitude



import os

import numpy as np
from numpy.testing import assert_array_equal

from audiomentations import (
    ClippingDistortion,
    AddBackgroundNoise,
    TimeMask,
    Shift,
    Compose,
)
from demo.demo import DEMO_DIR


class TestCompose:
    def test_freeze_and_unfreeze_parameters(self):
        samples = np.zeros((20,), dtype=np.float32)
        sample_rate = 44100
        augmenter = Compose(
            [
                AddBackgroundNoise(
                    sounds_path=os.path.join(DEMO_DIR, "background_noises"),
                    min_snr_db=15,
                    max_snr_db=35,
                    p=1.0,
                ),
                ClippingDistortion(p=0.5),
            ]
        )
        perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
        augmenter.freeze_parameters()
        for transform in augmenter.transforms:
            assert transform.are_parameters_frozen
        perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

        assert_array_equal(perturbed_samples1, perturbed_samples2)

        augmenter.unfreeze_parameters()
        for transform in augmenter.transforms:
            assert not transform.are_parameters_frozen

    def test_randomize_parameters_and_apply(self):
        samples = 1.0 / np.arange(1, 21, dtype=np.float32)
        sample_rate = 44100

        augmenter = Compose(
            [
                AddBackgroundNoise(
                    sounds_path=os.path.join(DEMO_DIR, "background_noises"),
                    min_snr_db=15,
                    max_snr_db=35,
                    p=1.0,
                ),
                ClippingDistortion(p=0.5),
                TimeMask(min_band_part=0.2, max_band_part=0.5, p=0.5),
                Shift(min_shift=0.5, max_shift=0.5, p=0.5),
            ]
        )
        augmenter.freeze_parameters()
        augmenter.randomize_parameters(samples=samples, sample_rate=sample_rate)

        parameters = [transform.parameters for transform in augmenter.transforms]

        perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
        perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

        assert_array_equal(perturbed_samples1, perturbed_samples2)

        augmenter.unfreeze_parameters()

        for transform_parameters, transform in zip(parameters, augmenter.transforms):
            assert transform_parameters == transform.parameters
            assert not transform.are_parameters_frozen



import numpy as np

from audiomentations import Trim, Compose


class TestTrim:
    def test_trim(self):
        sample_len = 1024
        samples1 = np.zeros((sample_len,), dtype=np.float32)
        samples2 = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000
        augmenter = Compose([Trim(top_db=20, p=1.0)])
        samples_in = np.hstack((samples1, samples2))
        assert len(samples_in) == sample_len * 2
        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)

        assert samples_out.dtype == np.float32
        assert len(samples_out) < sample_len * 2



import os
import pytest

import librosa
import numpy as np

from audiomentations import SpecChannelShuffle
from audiomentations.core.audio_loading_utils import load_sound_file
from audiomentations.core.transforms_interface import MonoAudioNotSupportedException
from demo.demo import DEMO_DIR
from .utils import plot_matrix

DEBUG = False


class TestSpecChannelShuffle:
    def test_shuffle_channels(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "background_noises", "hens.ogg"),
            sample_rate=None,
            mono=False,
        )
        assert samples.shape[0] == 2
        magnitude_spectrogram_chn0 = librosa.feature.melspectrogram(
            y=np.asfortranarray(samples[0, :]), sr=sample_rate
        )
        magnitude_spectrogram_chn1 = librosa.feature.melspectrogram(
            y=np.asfortranarray(samples[1, :]), sr=sample_rate
        )
        multichannel_magnitude_spectrogram = np.zeros(
            shape=(
                magnitude_spectrogram_chn0.shape[0],
                magnitude_spectrogram_chn0.shape[1],
                3,
            ),
            dtype=np.float32,
        )
        multichannel_magnitude_spectrogram[:, :, 0] = magnitude_spectrogram_chn0
        multichannel_magnitude_spectrogram[:, :, 1] = magnitude_spectrogram_chn1
        multichannel_magnitude_spectrogram[:, :, 2] = magnitude_spectrogram_chn1 * 0.7

        if DEBUG:
            image = (7 + np.log10(multichannel_magnitude_spectrogram + 0.0000001)) / 8
            plot_matrix(image, title="before")

        # Make sure the shuffled channels do not equal the original order
        transform = SpecChannelShuffle(p=1.0)
        for _ in range(100000):
            transform.randomize_parameters(multichannel_magnitude_spectrogram)
            if transform.parameters["shuffled_channel_indexes"] != [0, 1, 2]:
                break
        transform.freeze_parameters()

        augmented_spectrogram = transform(multichannel_magnitude_spectrogram)

        if DEBUG:
            image = (7 + np.log10(augmented_spectrogram + 0.0000001)) / 8
            plot_matrix(image, title="after")

        with np.testing.assert_raises(AssertionError):
            np.testing.assert_array_equal(
                augmented_spectrogram, multichannel_magnitude_spectrogram
            )

        for augmented_index, original_index in enumerate(
            transform.parameters.get("shuffled_channel_indexes")
        ):
            np.testing.assert_array_equal(
                augmented_spectrogram[:, :, augmented_index],
                multichannel_magnitude_spectrogram[:, :, original_index],
            )

    def test_shuffle_channels_mono(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "acoustic_guitar_0.wav"), sample_rate=None
        )
        magnitude_spectrogram = librosa.feature.melspectrogram(
            y=samples, sr=sample_rate
        )

        transform = SpecChannelShuffle(p=1.0)
        with pytest.raises(MonoAudioNotSupportedException):
            augmented_spectrogram = transform(magnitude_spectrogram)

    def test_empty_spectrogram(self):
        spec = np.zeros(shape=(0, 0), dtype=np.float32)
        transform = SpecChannelShuffle(p=1.0)
        augmented_spectrogram = transform(spec)

        np.testing.assert_array_equal(spec, augmented_spectrogram)



import os

import numpy as np
import pytest

from audiomentations.core.utils import (
    calculate_desired_noise_rms,
    convert_decibels_to_amplitude_ratio,
    find_audio_files_in_paths,
    calculate_rms,
    calculate_rms_without_silence,
    a_weighting_frequency_envelope,
)
from demo.demo import DEMO_DIR


class TestUtils:
    def test_calculate_desired_noise_rms(self):
        noise_rms = calculate_desired_noise_rms(clean_rms=0.5, snr=6)
        assert noise_rms == pytest.approx(0.2505936168136362)

    def test_convert_decibels_to_amplitude_ratio(self):
        amplitude_ratio = convert_decibels_to_amplitude_ratio(decibels=-6)
        assert amplitude_ratio == pytest.approx(0.5011872336272722)

        amplitude_ratio = convert_decibels_to_amplitude_ratio(decibels=6)
        assert amplitude_ratio == pytest.approx(1.9952623149688795)

    def test_find_audio_files_in_paths_uppercase_extension(self):
        file_paths = find_audio_files_in_paths(DEMO_DIR, traverse_subdirectories=False)
        found_it = False
        for file_path in file_paths:
            if file_path.name == "stereo_24bit.WAV":
                found_it = True
                break
        assert found_it

    def test_find_single_audio_file(self):
        file_paths = find_audio_files_in_paths(
            os.path.join(DEMO_DIR, "bus.opus"), traverse_subdirectories=False
        )
        assert len(file_paths) == 1
        assert file_paths[0].name == "bus.opus"

    def test_find_multiple_audio_files(self):
        file_paths = find_audio_files_in_paths(
            [os.path.join(DEMO_DIR, "bus.opus"), os.path.join(DEMO_DIR, "testing.m4a")]
        )
        assert len(file_paths) == 2
        assert file_paths[0].name == "bus.opus"
        assert file_paths[1].name == "testing.m4a"

    def test_find_audio_files_multiple_dirs(self):
        file_paths = find_audio_files_in_paths(
            [os.path.join(DEMO_DIR, "short_noises"), os.path.join(DEMO_DIR, "ir")]
        )
        assert len(file_paths) == 8
        assert file_paths[0].name == "130921_iPhone_rub_channel0_chunk83_aug2.wav"

    def test_calculate_rms_stereo(self):
        np.random.seed(42)
        sample_len = 1024
        samples_in = np.random.uniform(low=-0.5, high=0.5, size=(2, sample_len)).astype(
            np.float32
        )
        rms = calculate_rms(samples_in)
        assert rms == pytest.approx(0.287, abs=0.01)

    def test_calculate_rms_without_silence(self):
        sample_rate = 48000
        samples_in = np.zeros(int(2.0022 * sample_rate))
        samples_in[0:sample_rate] = 0.4 * np.ones(sample_rate)
        rms_before = calculate_rms(samples_in)
        rms_after = calculate_rms_without_silence(samples_in, sample_rate)
        assert rms_after > rms_before
        assert rms_after == pytest.approx(0.4)

        # Check that the function works if the input is shorter than a window (25 ms)
        rms_short = calculate_rms_without_silence(
            samples_in[0 : int(0.015 * sample_rate)], sample_rate
        )
        assert rms_short == pytest.approx(0.4)

    @pytest.mark.parametrize(
        "sample_rate, n_fft",
        [
            (16000, 1024),
            (48000, 4096),
        ],
    )
    def test_a_weighting_curve(self, sample_rate, n_fft):
        freqs = np.fft.rfftfreq(n_fft, 1 / sample_rate)
        weighting = (
            (12194**2 * freqs**4)
            / (
                (freqs**2 + 20.6**2)
                * np.sqrt((freqs**2 + 107.7**2) * (freqs**2 + 737.9**2))
                * (freqs**2 + 12194**2)
            )
        ) + 2.00

        assert a_weighting_frequency_envelope(n_fft, sample_rate) == pytest.approx(
            weighting, abs=0.01
        )



from copy import deepcopy

import numpy as np
from numpy.testing import assert_array_equal

from audiomentations import PitchShift, Compose


class TestPitchShift:
    def test_apply_pitch_shift(self):
        samples = np.zeros((2048,), dtype=np.float32)
        sample_rate = 16000
        augmenter = Compose([PitchShift(min_semitones=-2, max_semitones=-1, p=1.0)])
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape[-1] == 2048

    def test_apply_pitch_shift_multichannel(self):
        num_channels = 3
        samples = np.random.normal(0, 0.1, size=(num_channels, 5555)).astype(np.float32)
        sample_rate = 16000
        augmenter = Compose([PitchShift(min_semitones=1, max_semitones=2, p=1.0)])
        samples_out = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples.shape
        for i in range(num_channels):
            assert not np.allclose(samples[i], samples_out[i])

    def test_freeze_parameters(self):
        """
        Test that the transform can freeze its parameters, e.g. to apply the effect with the
        same parameters to multiple sounds.
        """
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 8000)).astype(np.float32)
        sample_rate = 16000
        augmenter = Compose([PitchShift(min_semitones=1, max_semitones=12, p=1.0)])

        first_samples = augmenter(samples=samples, sample_rate=sample_rate)
        first_parameters = deepcopy(augmenter.transforms[0].parameters)

        augmenter.transforms[0].min_semitones = -12
        augmenter.transforms[0].max_semitones = -1
        augmenter.transforms[0].are_parameters_frozen = True
        second_samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert first_parameters == augmenter.transforms[0].parameters
        assert_array_equal(first_samples, second_samples)



import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class AdjustDuration(BaseWaveformTransform):
    """
    Trim or pad the audio to the specified length/duration in samples or seconds. If the
    input sound is longer than the target duration, pick a random offset and crop the
    sound to the target duration. If the input sound is shorter than the target
    duration, pad the sound so the duration matches the target duration.
    """

    supports_multichannel = True

    def __init__(
        self,
        duration_samples: int = None,
        duration_seconds: float = None,
        padding_mode: str = "silence",
        padding_position: str = "end",
        p: float = 0.5,
    ):
        """
        :param duration_samples: Target duration in number of samples
        :param duration_seconds: Target duration in seconds
        :param padding_mode: Padding mode. Must be "silence", "wrap" or "reflect". Only
            used when audio input is shorter than the target duration.
        :param padding_position: The position of the inserted/added padding. Must be
            "start" or "end". Only used when audio input is shorter than the target duration.
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert padding_mode in ("silence", "wrap", "reflect")
        if padding_mode == "silence":
            padding_mode = "constant"  # for numpy.pad compatibility
        self.padding_mode = padding_mode

        assert padding_position in ("start", "end")
        self.padding_position = padding_position
        
        assert duration_samples is not None or duration_seconds is not None
        if duration_samples is not None and duration_seconds is not None:
            raise ValueError(
                "should have duration_samples or duration_seconds, but not both"
            )
        elif duration_seconds:
            assert duration_seconds > 0
            self.get_target_samples = lambda sr: int(duration_seconds * sr)
        elif duration_samples:
            assert duration_samples > 0
            self.get_target_samples = lambda sr: duration_samples
            
        self.duration_samples = duration_samples
        self.duration_seconds = duration_seconds

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        target_samples = self.get_target_samples(sample_rate)
        sample_length = samples.shape[-1]

        if sample_length == target_samples:
            return samples

        elif sample_length > target_samples:
            start = np.random.randint(0, sample_length - target_samples)
            return samples[..., start : start + target_samples]

        elif sample_length < target_samples:
            padding_length = target_samples - sample_length
            if samples.ndim == 1:
                if self.padding_position == "start":
                    pad_width = (padding_length, 0)
                else:
                    pad_width = (0, padding_length)
            else:
                if self.padding_position == "start":
                    pad_width = ((0, 0), (padding_length, 0))
                else:
                    pad_width = ((0, 0), (0, padding_length))
            return np.pad(samples, pad_width, self.padding_mode)



import numpy as np
import pytest

from audiomentations import Clip


class TestClip:
    def test_single_channel(self):
        samples = np.array([0.5, 0.6, -0.2, 0.0], dtype=np.float32)
        sample_rate = 16000
        augmenter = Clip(a_min=-0.1, a_max=0.1, p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert np.amin(samples) == pytest.approx(-0.1)
        assert np.amax(samples) == pytest.approx(0.1)
        assert samples.dtype == np.float32
        assert samples.shape[-1] == 4

    def test_multichannel(self):
        samples = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        sample_rate = 16000
        augmenter = Clip(a_min=-0.1, a_max=0.1, p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert np.amin(samples) == pytest.approx(-0.1)
        assert np.amax(samples) == pytest.approx(0.1)
        assert samples.dtype == np.float32



import math
import os
import pytest
import warnings

import numpy as np

from audiomentations.core.audio_loading_utils import (
    load_sound_file,
)
from demo.demo import DEMO_DIR


class TestLoadSoundFiles:
    def test_load_stereo_ogg_vorbis(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "background_noises", "hens.ogg"), sample_rate=None
        )
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        # Apparently, the exact duration may vary slightly based on which decoder is used
        assert samples.shape[0] >= 442575
        assert samples.shape[0] <= 443328

        max_value = np.amax(samples)
        assert max_value > 0.02
        assert max_value < 1.0

    def test_load_mono_opus(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "bus.opus"), sample_rate=None
        )
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        # Apparently, the exact duration may vary slightly based on which decoder is used
        assert samples.shape[0] >= 36682
        assert samples.shape[0] <= 36994

        max_value = np.amax(samples)
        assert max_value > 0.3
        assert max_value < 1.0

    def test_load_mono_m4a(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "testing.m4a"), sample_rate=None
        )
        assert sample_rate == 44100
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        max_value = np.amax(samples)
        assert max_value > 0.1
        assert max_value < 1.0

    def test_load_mono_signed_16_bit_wav(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "acoustic_guitar_0.wav"), sample_rate=None
        )
        assert sample_rate == 16000
        assert samples.dtype == np.float32
        assert samples.ndim == 1
        
        assert samples.shape[0] == 140544

        max_value = np.amax(samples)
        assert max_value > 0.5
        assert max_value < 1.0

    def test_load_stereo_signed_16_bit_wav(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "stereo_16bit.wav"), sample_rate=None
        )
        assert sample_rate == 16000
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[0] == 17833

        max_value = np.amax(samples)
        assert max_value > 0.5
        assert max_value < 1.0

    def test_load_mono_signed_24_bit_wav(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "signed_24bit.wav"), sample_rate=None
        )
        assert sample_rate == 48000
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[0] == 54514

        max_value = np.amax(samples)
        assert max_value > 0.09
        assert max_value < 1.0

    def test_load_mono_signed_24_bit_wav2(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "mono_int24.wav"), sample_rate=None
        )
        assert sample_rate == 44100
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[-1] == 22

        max_value = np.amax(samples)
        assert max_value == pytest.approx(0.96750367)
        min_value = np.amin(samples)
        assert min_value == pytest.approx(-0.9822748)

    def test_load_mono_signed_32_bit_wav(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "mono_int32.wav"), sample_rate=None
        )
        assert sample_rate == 44100
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[-1] == 22

        max_value = np.amax(samples)
        assert max_value == pytest.approx(0.96750367)
        min_value = np.amin(samples)
        assert min_value == pytest.approx(-0.9822748)

    def test_load_mono_float64_wav(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "mono_float64.wav"), sample_rate=None
        )
        assert sample_rate == 44100
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[-1] == 22

        max_value = np.amax(samples)
        assert max_value == pytest.approx(0.96750367)
        min_value = np.amin(samples)
        assert min_value == pytest.approx(-0.9822748)

    def test_load_stereo_signed_24_bit_wav(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "stereo_24bit.WAV"), sample_rate=None
        )
        assert sample_rate == 16000
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[0] == 17833

        max_value = np.amax(samples)
        assert max_value > 0.5
        assert max_value < 1.0
        
    def test_load_mono_ms_adpcm(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "ms_adpcm.wav"), sample_rate=None
        )
        assert sample_rate == 11024
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[0] == 895500

        max_value = np.amax(samples)
        assert max_value > 0.3
        assert max_value < 1.0

    def test_load_mono_ms_adpcm_and_resample(self):
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            samples, sample_rate = load_sound_file(
                os.path.join(DEMO_DIR, "ms_adpcm.wav"), sample_rate=16000
            )

            assert len(w) >= 1

        assert sample_rate == 16000
        assert samples.dtype == np.float32
        assert samples.ndim == 1

        assert samples.shape[0] == math.ceil(895500 * 16000 / 11024)

        max_value = np.amax(samples)
        assert max_value > 0.3
        assert max_value < 1.0



import json
import random

import fast_align_audio
import numpy as np
import pytest

from audiomentations import Limiter


class TestLimiter:
    @pytest.mark.parametrize(
        "samples_in",
        [
            np.random.normal(0, 1, size=1000).astype(np.float32),
            np.random.normal(0, 0.001, size=(1, 250)).astype(np.float32),
            np.random.normal(0, 0.1, size=(3, 8888)).astype(np.float32),
        ],
    )
    def test_limiter(self, samples_in):
        augmenter = Limiter(p=1.0, min_attack=0.0025, max_attack=0.0025)
        std_in = np.mean(np.abs(samples_in))
        samples_out = augmenter(samples=samples_in, sample_rate=16000)
        std_out = np.mean(np.abs(samples_out))
        length = samples_in.shape[-1]

        samples_in_mono = samples_in
        samples_out_mono = samples_out
        if samples_in_mono.ndim > 1:
            samples_in_mono = samples_in_mono[0]
            samples_out_mono = samples_out_mono[0]
        offset, _ = fast_align_audio.find_best_alignment_offset(
            reference_signal=samples_in_mono,
            delayed_signal=samples_out_mono,
            max_offset_samples=length // 2,
            lookahead_samples=length // 2,
        )
        # Check that the output is aligned with the input, i.e. no delay was introduced
        assert offset == 0

        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape
        assert std_out < std_in

    def test_limiter_validation(self):
        with pytest.raises(AssertionError):
            Limiter(min_attack=-0.5)

    def test_serialize_parameters(self):
        random.seed(42)
        transform = Limiter(p=1.0)
        samples = np.random.normal(0, 1, size=1024).astype(np.float32)
        transform.randomize_parameters(samples, sample_rate=16000)
        json.dumps(transform.serialize_parameters())

    def test_digital_silence(self):
        samples_in = np.zeros((1024,), np.float32)
        augmenter = Limiter(p=1.0)
        std_in = np.mean(np.abs(samples_in))
        samples_out = augmenter(samples=samples_in, sample_rate=16000)
        std_out = np.mean(np.abs(samples_out))
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape
        assert std_out == std_in == 0.0



import numpy as np
import pytest

from audiomentations import AddGaussianNoise, Compose


class TestGaussianNoise:
    def test_gaussian_noise(self):
        samples = np.zeros((20,), dtype=np.float32)
        sample_rate = 16000
        augmenter = Compose([AddGaussianNoise(p=1.0)])
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert not (float(np.sum(np.abs(samples))) == pytest.approx(0.0))

    def test_gaussian_noise_stereo(self):
        samples = np.zeros((2, 2000), dtype=np.float32)
        sample_rate = 16000
        augmenter = Compose([AddGaussianNoise(p=1.0)])
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert not (float(np.sum(np.abs(samples))) == pytest.approx(0.0))



import os

import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

from audiomentations import (
    ClippingDistortion,
    AddBackgroundNoise,
    TimeMask,
    Shift,
    PolarityInversion,
    Gain,
    SpecFrequencyMask,
    OneOf,
)
from demo.demo import DEMO_DIR


class TestOneOf:
    def test_exactly_one_transform_applied(self):
        samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
        sample_rate = 44100

        for _ in range(30):
            augmenter = OneOf(
                [
                    Gain(min_gain_db=-12, max_gain_db=-6, p=1.0),
                    PolarityInversion(p=1.0),
                ]
            )
            perturbed_samples = augmenter(samples=samples, sample_rate=sample_rate)

            was_gain_applied = 0.0 < abs(perturbed_samples[0]) < 0.25
            was_polarity_inversion_applied = perturbed_samples[0] < 0.0

            num_transforms_applied = sum(
                [
                    1 if was_gain_applied else 0,
                    1 if was_polarity_inversion_applied else 0,
                ]
            )
            assert num_transforms_applied == 1

    def test_freeze_and_unfreeze_all_parameters(self):
        samples = np.array([0.25, 0.0, 0.1, -0.4], dtype=np.float32)
        sample_rate = 44100

        for _ in range(30):
            augmenter = OneOf(
                [
                    Gain(p=1.0),
                    PolarityInversion(p=1.0),
                ]
            )
            perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
            augmenter.freeze_parameters()
            for transform in augmenter.transforms:
                assert transform.are_parameters_frozen
            perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

            assert_array_equal(perturbed_samples1, perturbed_samples2)

            augmenter.unfreeze_parameters()
            for transform in augmenter.transforms:
                assert not transform.are_parameters_frozen

    def test_freeze_and_unfreeze_own_parameters(self):
        augmenter = OneOf(
            [
                Gain(p=1.0),
                PolarityInversion(p=1.0),
            ]
        )
        assert not augmenter.are_parameters_frozen
        for transform in augmenter.transforms:
            assert not transform.are_parameters_frozen

        augmenter.freeze_parameters(apply_to_children=False)
        assert augmenter.are_parameters_frozen
        for transform in augmenter.transforms:
            assert not transform.are_parameters_frozen

        augmenter.unfreeze_parameters(apply_to_children=False)
        assert not augmenter.are_parameters_frozen
        for transform in augmenter.transforms:
            assert not transform.are_parameters_frozen

    def test_randomize_all_parameters_and_apply(self):
        samples = 1.0 / np.arange(1, 21, dtype=np.float32)
        sample_rate = 44100

        augmenter = OneOf(
            [
                AddBackgroundNoise(
                    sounds_path=os.path.join(DEMO_DIR, "background_noises"),
                    min_snr_db=15,
                    max_snr_db=35,
                    p=1.0,
                ),
                ClippingDistortion(p=1.0),
                TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0),
                Shift(min_shift=0.5, max_shift=0.5, p=1.0),
            ]
        )
        augmenter.freeze_parameters()
        augmenter.randomize_parameters(samples=samples, sample_rate=sample_rate)

        parameters = [transform.parameters for transform in augmenter.transforms]

        perturbed_samples1 = augmenter(samples=samples, sample_rate=sample_rate)
        perturbed_samples2 = augmenter(samples=samples, sample_rate=sample_rate)

        assert_array_equal(perturbed_samples1, perturbed_samples2)

        augmenter.unfreeze_parameters()

        for transform_parameters, transform in zip(parameters, augmenter.transforms):
            assert transform_parameters == transform.parameters
            assert not transform.are_parameters_frozen

    def test_randomize_only_own_parameters(self):
        samples = 1.0 / np.arange(1, 21, dtype=np.float32)
        sample_rate = 44100

        augmenter = OneOf(
            [
                Gain(p=1.0),
                PolarityInversion(p=1.0),
            ]
        )
        augmenter.randomize_parameters(samples, sample_rate, apply_to_children=True)
        own_parameters_before = (augmenter.should_apply, augmenter.transform_index)

        parameters = [transform.parameters for transform in augmenter.transforms]

        own_parameters_changed = False
        for i in range(30):
            augmenter.randomize_parameters(
                samples, sample_rate, apply_to_children=False
            )
            if (
                augmenter.should_apply,
                augmenter.transform_index,
            ) != own_parameters_before:
                own_parameters_changed = True
                break
        assert own_parameters_changed

        # Check that the children's parameters are still the same
        for transform_parameters, transform in zip(parameters, augmenter.transforms):
            assert transform_parameters == transform.parameters
            assert not transform.are_parameters_frozen

    def test_one_of_spectrogram_magnitude(self):
        spectrogram = np.random.random((128, 128, 2))
        augmenter = OneOf(
            [
                SpecFrequencyMask(fill_mode="mean", p=1.0),
                SpecFrequencyMask(fill_mode="constant", p=1.0),
            ]
        )

        # Positional argument
        augmented_spectrogram = augmenter(spectrogram)
        assert augmented_spectrogram.shape == spectrogram.shape
        assert augmented_spectrogram.dtype == spectrogram.dtype
        with np.testing.assert_raises(AssertionError):
            assert_array_almost_equal(augmented_spectrogram, spectrogram)

        # Keyword argument
        augmented_spectrogram = augmenter(magnitude_spectrogram=spectrogram)
        assert augmented_spectrogram.shape == spectrogram.shape
        assert augmented_spectrogram.dtype == spectrogram.dtype
        with np.testing.assert_raises(AssertionError):
            assert_array_almost_equal(augmented_spectrogram, spectrogram)

    def test_one_of_spectrogram_magnitude_with_p_0(self):
        spectrogram = np.random.random((128, 128, 2))
        augmenter = OneOf(
            [
                SpecFrequencyMask(fill_mode="mean", p=1.0),
                SpecFrequencyMask(fill_mode="constant", p=1.0),
            ],
            p=0.0,
        )

        # Positional argument
        augmented_spectrogram = augmenter(spectrogram)
        assert augmented_spectrogram.shape == spectrogram.shape
        assert augmented_spectrogram.dtype == spectrogram.dtype
        assert_array_equal(augmented_spectrogram, spectrogram)

        # Keyword argument
        augmented_spectrogram = augmenter(magnitude_spectrogram=spectrogram)
        assert augmented_spectrogram.shape == spectrogram.shape
        assert augmented_spectrogram.dtype == spectrogram.dtype
        assert_array_equal(augmented_spectrogram, spectrogram)



import librosa
import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


def next_power_of_2(x: int) -> int:
    """
    taken jhoyla's answer here:
    https://stackoverflow.com/questions/14267555/find-the-smallest-power-of-2-greater-than-or-equal-to-n-in-python
    """
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


def get_temperature_humidity_key(temperature: float, humidity: float) -> str:
    key = str(int(temperature)) + "C"
    bounds = [30, 50, 70, 90]
    for n in range(1, len(bounds)):
        if bounds[n - 1] <= humidity <= bounds[n]:
            key += f"_{bounds[n-1]}-{bounds[n]}%"
            break
    return key


class AirAbsorption(BaseWaveformTransform):
    """
    Apply a Lowpass-like filterbank with variable octave attenuation that simulates attenuation of
    higher frequencies due to air absorption. This transform is parametrized by temperature,
    humidity, and the distance between audio source and microphone.

    This is not a scientifically accurate transform but basically applies a uniform
    filterbank with attenuations given by:

    att = exp(- distance * absorption_coefficient)

    where distance is the microphone-source assumed distance in meters and `absorption_coefficient`
    is adapted from a lookup table by pyroomacoustics [1]. It can also be seen as a lowpass filter
    with variable octave attenuation.

    Note: This only "simulates" the damping of high frequencies, and does not
    attenuate according to the distance law. Gain augmentation needs to be done separately.

    [1] https://github.com/LCAV/pyroomacoustics
    """

    supports_multichannel = True

    # Table of air absorption coefficients adapted from `pyroomacoustics`.
    # The keys are of the form:
    #   "<degrees>C_<minimum_humidity>-<maximum_humidity>%"
    #
    # And the values are attenuation coefficients `coef` that attenuate the corresponding band
    # in "center_freq" by exp(-coef * <microphone-source distance>).
    # The original table does not have the last two columns which have been extrapolated from the
    # pyroomacoustics table using `curve_fit`
    air_absorption_table = {
        "10C_30-50%": [
            x * 1e-3 for x in [0.1, 0.2, 0.5, 1.1, 2.7, 9.4, 29.0, 91.5, 289.0]
        ],
        "10C_50-70%": [
            x * 1e-3 for x in [0.1, 0.2, 0.5, 0.8, 1.8, 5.9, 21.1, 76.6, 280.2]
        ],
        "10C_70-90%": [
            x * 1e-3 for x in [0.1, 0.2, 0.5, 0.7, 1.4, 4.4, 15.8, 58.0, 214.9]
        ],
        "20C_30-50%": [
            x * 1e-3 for x in [0.1, 0.3, 0.6, 1.0, 1.9, 5.8, 20.3, 72.3, 259.9]
        ],
        "20C_50-70%": [
            x * 1e-3 for x in [0.1, 0.3, 0.6, 1.0, 1.7, 4.1, 13.5, 44.4, 148.7]
        ],
        "20C_70-90%": [
            x * 1e-3 for x in [0.1, 0.3, 0.6, 1.1, 1.7, 3.5, 10.6, 31.2, 93.8]
        ],
        "center_freqs": [125, 250, 500, 1000, 2000, 4000, 8000, 16000, 32000],
    }

    def __init__(
        self,
        min_temperature: float = 10.0,
        max_temperature: float = 20.0,
        min_humidity: float = 30.0,
        max_humidity: float = 90.0,
        min_distance: float = 10.0,
        max_distance: float = 100.0,
        p=0.5,
    ):
        """
        :param min_temperature: Minimum temperature in Celsius (can take a value of either 10.0 or 20.0)
        :param max_temperature: Maximum temperature in Celsius (can take a value of either 10.0 or 20.0)
        :param min_humidity: Minimum humidity in percent (between 30 and 90)
        :param max_humidity: Maximum humidity in percent (between 30 and 90)
        :param min_distance: Minimum microphone-source distance in meters.
        :param max_distance: Maximum microphone-source distance in meters.
        :param p: The probability of applying this transform
        """
        assert float(min_temperature) in [
            10.0,
            20.0,
        ], "Sorry, the only supported temperatures are either 10 or 20 degrees Celsius"
        assert float(max_temperature) in [
            10.0,
            20.0,
        ], "Sorry, the only supported temperatures are either 10 or 20 degrees Celsius"
        assert min_temperature <= max_temperature
        assert 30 <= min_humidity <= max_humidity <= 90
        assert min_distance > 0.0
        assert max_distance > 0.0
        assert min_distance <= max_distance

        super().__init__(p)

        self.min_temperature = min_temperature
        self.max_temperature = max_temperature

        self.min_humidity = min_humidity
        self.max_humidity = max_humidity

        self.min_distance = min_distance
        self.max_distance = max_distance

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        self.parameters["temperature"] = 10 * np.random.randint(
            int(self.min_temperature) // 10, int(self.max_temperature) // 10 + 1
        )
        self.parameters["humidity"] = np.random.randint(
            self.min_humidity, self.max_humidity + 1
        )
        self.parameters["distance"] = np.random.uniform(
            self.min_distance, self.max_distance
        )

    def apply(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        assert samples.dtype == np.float32

        # Choose correct absorption coefficients
        key = get_temperature_humidity_key(
            self.parameters["temperature"], self.parameters["humidity"]
        )

        # Convert to attenuations
        attenuation_values = np.exp(
            -self.parameters["distance"] * np.array(self.air_absorption_table[key])
        )

        # Calculate n_fft so that the lowest band can be stored in a single
        # fft bin.
        first_band_bw = self.air_absorption_table["center_freqs"][0] / (2**0.5)
        n_fft = next_power_of_2(int(sample_rate / 2 / first_band_bw))

        # Frequencies to calculate the attenuations caused by air absorption
        frequencies = librosa.fft_frequencies(sr=sample_rate, n_fft=n_fft)

        # Interpolate to the desired frequencies (we have to do this in dB)
        db_target_attenuations = np.interp(
            frequencies,
            self.air_absorption_table["center_freqs"],
            20 * np.log10(attenuation_values),
        )

        linear_target_attenuations = 10 ** (db_target_attenuations / 20)

        # Apply using STFT
        if len(samples.shape) == 1:
            stft = librosa.stft(samples, n_fft=n_fft)

            # Compute mask
            mask = np.tile(linear_target_attenuations, (stft.shape[1], 1)).T

            # Compute target degraded audio
            result = librosa.istft(stft * mask, length=len(samples), dtype=np.float32)

        else:
            result = np.zeros_like(samples, dtype=np.float32)

            for chn_idx, channel in enumerate(samples):
                stft = librosa.stft(channel, n_fft=n_fft)

                # Compute mask
                mask = np.tile(linear_target_attenuations, (stft.shape[1], 1)).T

                # Compute target degraded audio
                result[chn_idx, :] = librosa.istft(stft * mask, length=result.shape[1])

        return result



import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

from audiomentations import Mp3Compression
from audiomentations.core.audio_loading_utils import load_sound_file
from demo.demo import DEMO_DIR


def fast_autocorr(original, delayed, t=1):
    if t == 0:
        return np.corrcoef([original[::4], delayed[::4]])[1, 0]
    elif t < 0:
        return np.corrcoef([original[-t::4], delayed[:t:4]])[1, 0]
    else:
        return np.corrcoef([original[:-t:4], delayed[t::4]])[1, 0]


def calculate_delay(
    original_signal,
    delayed_signal,
    min_lag_in_samples,
    max_lag_in_samples,
    debug=False,
):
    coefs = []
    for lag in range(min_lag_in_samples, max_lag_in_samples):
        correlation_coef = fast_autocorr(original_signal, delayed_signal, t=lag)
        coefs.append(correlation_coef)

    max_coef_index = int(np.argmax(coefs))

    if coefs[max_coef_index] < 0.3:
        print(
            "Warning: Max coefficient is low ({})! Please check if the delay is out of bounds".format(
                coefs[max_coef_index]
            )
        )
    delay = max_coef_index + min_lag_in_samples

    if debug:
        plt.plot(coefs)
        plt.show()

    return delay


def mse(signal1, signal2):
    """Mean squared error"""
    return np.mean((signal1 - signal2) ** 2)


class TestMp3Compression:
    @pytest.mark.parametrize(
        "backend",
        ["pydub", "lameenc"],
    )
    @pytest.mark.parametrize(
        "shape",
        [(44100,), (1, 22049), (2, 10000)],
    )
    def test_apply_mp3_compression(self, backend: str, shape: tuple):
        samples_in = np.random.normal(0, 1, size=shape).astype(np.float32)
        sample_rate = 44100
        augmenter = Mp3Compression(
            p=1.0, min_bitrate=48, max_bitrate=48, backend=backend
        )

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert len(shape) == len(samples_out.shape)
        assert samples_out.dtype == np.float32
        assert samples_out.shape[-1] >= shape[-1]
        assert samples_out.shape[-1] < shape[-1] + 3000
        if len(shape) == 2:
            assert samples_out.shape[0] == shape[0]

    @pytest.mark.parametrize(
        "backend",
        ["pydub", "lameenc"],
    )
    @pytest.mark.parametrize(
        "shape",
        [(16000,), (1, 12049), (2, 5000)],
    )
    def test_apply_mp3_compression_low_bitrate(self, backend: str, shape: tuple):
        samples_in = np.random.normal(0, 1, size=shape).astype(np.float32)
        sample_rate = 16000
        augmenter = Mp3Compression(p=1.0, min_bitrate=8, max_bitrate=8, backend=backend)

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert len(shape) == len(samples_out.shape)
        assert samples_out.dtype == np.float32
        assert samples_out.shape[-1] >= shape[-1]
        assert samples_out.shape[-1] < shape[-1] + 3100
        if len(shape) == 2:
            assert samples_out.shape[0] == shape[0]

    def test_invalid_argument_combination(self):
        with pytest.raises(ValueError):
            _ = Mp3Compression(min_bitrate=400, max_bitrate=800)

        with pytest.raises(ValueError):
            _ = Mp3Compression(min_bitrate=2, max_bitrate=128)

        with pytest.raises(ValueError):
            _ = Mp3Compression(min_bitrate=64, max_bitrate=32)

        with pytest.raises(ValueError):
            _ = Mp3Compression(min_bitrate=66, max_bitrate=67)

        with pytest.raises(ValueError):
            _ = Mp3Compression(backend="both")

    def test_too_loud_input(self):
        """Check that we avoid wrap distortion if input is too loud"""
        samples_in, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "perfect-alley1.ogg"),
            sample_rate=None,
            mono=True,
        )
        samples_in = samples_in[..., 0:196000]
        samples_in *= 10
        augmenter = Mp3Compression(
            min_bitrate=320, max_bitrate=320, backend="lameenc", p=1.0
        )

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        samples_out = samples_out[..., 0 : samples_in.shape[-1]]
        assert samples_out.dtype == np.float32

        # Apply delay compensation, so we can compare output with input sample by sample
        delay = calculate_delay(
            samples_in,
            samples_out,
            min_lag_in_samples=0,
            max_lag_in_samples=3100,
        )
        snippet_length = 42000
        delay_compensated_samples_out = samples_out[..., delay:delay+snippet_length]
        samples_in = samples_in[..., :snippet_length]
        mse_value = mse(samples_in, delay_compensated_samples_out)

        assert mse_value < 0.00001



import random
from typing import Union

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import get_crossfade_mask_pair

# 0.00025 seconds corresponds to 2 samples at 8000 Hz
DURATION_EPSILON = 0.00025


class Shift(BaseWaveformTransform):
    """
    Shift the samples forwards or backwards, with or without rollover
    """

    supports_multichannel = True

    def __init__(
        self,
        min_shift: Union[float, int] = -0.5,
        max_shift: Union[float, int] = 0.5,
        shift_unit: str = "fraction",
        rollover: bool = True,
        fade_duration: float = 0.005,
        p: float = 0.5,
    ):
        """
        :param min_shift: Minimum amount of shifting in time. See also shift_unit.
        :param max_shift: Maximum amount of shifting in time. See also shift_unit.
        :param shift_unit: Defines the unit of the value of min_shift and max_shift.
            "fraction": Fraction of the total sound length
            "samples": Number of audio samples
            "seconds": Number of seconds
        :param rollover: When set to True, samples that roll beyond the
            last position are re-introduced at the first position, and samples that roll beyond the first
            position are re-introduced at the last position. When set to False, samples that roll beyond
            the first or last position are discarded. In other words, rollover=False results in
            an empty space (with zeroes).
        :param fade_duration: If you set this to a positive number (in seconds), there
            will be a fade in and/or out at the "stitch" (that was the start or the end
            of the audio before the shift). This can smooth out an unwanted abrupt
            change between two consecutive samples, which would otherwise sound like a transient/click/pop.
            This parameter denotes the duration of the fade in
            seconds. To disable the fading feature, set this parameter to 0.0.
        :param p: The probability of applying this transform
        """
        super().__init__(p)

        if min_shift > max_shift:
            raise ValueError("min_shift must not be greater than max_shift")
        if shift_unit not in ("fraction", "samples", "seconds"):
            raise ValueError('shift_unit must be "fraction", "samples" or "seconds"')
        if shift_unit == "fraction":
            if min_shift < -1:
                raise ValueError(
                    'min_shift must be >= -1 when shift_unit is "fraction"'
                )
            if max_shift > 1:
                raise ValueError('max_shift must be <= 1 when shift_unit is "fraction"')

        if fade_duration == 0.0:
            self.fade = False
        elif fade_duration < 0.0:
            raise ValueError("fade_duration must not be negative")
        elif fade_duration < DURATION_EPSILON:
            raise ValueError(
                "When fade_duration is set to a positive number, it must be >="
                f" {DURATION_EPSILON}"
            )
        else:
            self.fade = True

        self.min_shift = min_shift
        self.max_shift = max_shift
        self.shift_unit = shift_unit
        self.rollover = rollover
        self.fade_duration = fade_duration

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            if self.shift_unit == "samples":
                self.parameters["shift_amount"] = random.randint(
                    self.min_shift, self.max_shift
                )
            else:
                self.parameters["shift_amount"] = random.uniform(
                    self.min_shift, self.max_shift
                )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        if self.shift_unit == "samples":
            num_places_to_shift = self.parameters["shift_amount"]
        elif self.shift_unit == "fraction":
            num_places_to_shift = int(
                round(self.parameters["shift_amount"] * samples.shape[-1])
            )
        elif self.shift_unit == "seconds":
            num_places_to_shift = int(
                round(self.parameters["shift_amount"] * sample_rate)
            )
        else:
            raise ValueError("invalid shift_unit")

        shifted_samples = np.roll(samples, num_places_to_shift, axis=-1)

        if not self.rollover:
            if num_places_to_shift > 0:
                shifted_samples[..., :num_places_to_shift] = 0.0
            elif num_places_to_shift < 0:
                shifted_samples[..., num_places_to_shift:] = 0.0

        if self.fade:
            fade_length = int(sample_rate * self.fade_duration)
            fade_in, fade_out = get_crossfade_mask_pair(fade_length)

            if num_places_to_shift > 0:
                fade_in_start = num_places_to_shift
                fade_in_end = min(
                    num_places_to_shift + fade_length, shifted_samples.shape[-1]
                )
                fade_in_length = fade_in_end - fade_in_start

                shifted_samples[
                    ...,
                    fade_in_start:fade_in_end,
                ] *= fade_in[:fade_in_length]

                if self.rollover:
                    fade_out_start = max(num_places_to_shift - fade_length, 0)
                    fade_out_end = num_places_to_shift
                    fade_out_length = fade_out_end - fade_out_start

                    shifted_samples[..., fade_out_start:fade_out_end] *= fade_out[
                        -fade_out_length:
                    ]

            elif num_places_to_shift < 0:
                positive_num_places_to_shift = (
                    shifted_samples.shape[-1] + num_places_to_shift
                )

                fade_out_start = max(positive_num_places_to_shift - fade_length, 0)
                fade_out_end = positive_num_places_to_shift
                fade_out_length = fade_out_end - fade_out_start

                shifted_samples[..., fade_out_start:fade_out_end] *= fade_out[
                    -fade_out_length:
                ]

                if self.rollover:
                    fade_in_start = positive_num_places_to_shift
                    fade_in_end = min(
                        positive_num_places_to_shift + fade_length,
                        shifted_samples.shape[-1],
                    )
                    fade_in_length = fade_in_end - fade_in_start
                    shifted_samples[
                        ...,
                        fade_in_start:fade_in_end,
                    ] *= fade_in[:fade_in_length]

        return shifted_samples



import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import get_max_abs_amplitude


class Normalize(BaseWaveformTransform):
    """
    Apply a constant amount of gain, so that highest signal level present in the sound becomes
    0 dBFS, i.e. the loudest level allowed if all samples must be between -1 and 1. Also known
    as peak normalization.
    """

    supports_multichannel = True

    def __init__(self, apply_to: str = "all", p: float = 0.5):
        super().__init__(p)
        assert apply_to in ("all", "only_too_loud_sounds")
        self.apply_to = apply_to

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["max_amplitude"] = get_max_abs_amplitude(samples)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        if (
            self.apply_to == "only_too_loud_sounds"
            and self.parameters["max_amplitude"] < 1.0
        ):
            return samples

        if self.parameters["max_amplitude"] > 0:
            return samples / self.parameters["max_amplitude"]
        else:
            return samples



import os

import librosa
import numpy as np

from audiomentations import SpecFrequencyMask
from audiomentations.core.audio_loading_utils import load_sound_file
from demo.demo import DEMO_DIR
from .utils import plot_matrix

DEBUG = False


class TestSpecFrequencyMask:
    def test_fill_zeros(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "acoustic_guitar_0.wav"), sample_rate=None
        )
        magnitude_spectrogram = librosa.feature.melspectrogram(
            y=samples, sr=sample_rate
        )

        mask_fraction = 0.05
        transform = SpecFrequencyMask(
            fill_mode="constant",
            fill_constant=0.0,
            min_mask_fraction=mask_fraction,
            max_mask_fraction=mask_fraction,
            p=1.0,
        )
        augmented_spectrogram = transform(magnitude_spectrogram)

        if DEBUG:
            plot_matrix(np.log(augmented_spectrogram))

        with np.testing.assert_raises(AssertionError):
            np.testing.assert_array_equal(augmented_spectrogram, magnitude_spectrogram)

        num_zeroed_frequencies = 0
        for i in range(augmented_spectrogram.shape[0]):
            if sum(augmented_spectrogram[i]) == 0.0:
                num_zeroed_frequencies += 1

        assert num_zeroed_frequencies == int(
            round(magnitude_spectrogram.shape[0] * mask_fraction)
        )

    def test_fill_zeros_multichannel(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "background_noises", "hens.ogg"),
            sample_rate=None,
            mono=False,
        )
        assert samples.shape[0] == 2
        magnitude_spectrogram_chn0 = librosa.feature.melspectrogram(
            y=np.asfortranarray(samples[0, :]), sr=sample_rate
        )
        magnitude_spectrogram_chn1 = librosa.feature.melspectrogram(
            y=np.asfortranarray(samples[1, :]), sr=sample_rate
        )
        multichannel_magnitude_spectrogram = np.zeros(
            shape=(
                magnitude_spectrogram_chn0.shape[0],
                magnitude_spectrogram_chn0.shape[1],
                3,
            ),
            dtype=np.float32,
        )
        multichannel_magnitude_spectrogram[:, :, 0] = magnitude_spectrogram_chn0
        multichannel_magnitude_spectrogram[:, :, 1] = magnitude_spectrogram_chn1
        multichannel_magnitude_spectrogram[:, :, 2] = magnitude_spectrogram_chn1

        mask_fraction = 0.05
        transform = SpecFrequencyMask(
            fill_mode="constant",
            fill_constant=0.0,
            min_mask_fraction=mask_fraction,
            max_mask_fraction=mask_fraction,
            p=1.0,
        )
        augmented_spectrogram = transform(multichannel_magnitude_spectrogram)

        if DEBUG:
            image = (7 + np.log10(augmented_spectrogram + 0.0000001)) / 8
            plot_matrix(image)

        with np.testing.assert_raises(AssertionError):
            np.testing.assert_array_equal(
                augmented_spectrogram, multichannel_magnitude_spectrogram
            )

        num_zeroed_frequencies = 0
        for i in range(augmented_spectrogram.shape[0]):
            if np.sum(augmented_spectrogram[i]) == 0.0:
                num_zeroed_frequencies += 1

        assert num_zeroed_frequencies == int(
            round(multichannel_magnitude_spectrogram.shape[0] * mask_fraction)
        )

    def test_fill_mean(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "acoustic_guitar_0.wav"), sample_rate=None
        )
        magnitude_spectrogram = librosa.feature.melspectrogram(
            y=samples, sr=sample_rate
        )

        min_mask_fraction = 0.05
        max_mask_fraction = 0.09
        transform = SpecFrequencyMask(
            fill_mode="mean",
            min_mask_fraction=min_mask_fraction,
            max_mask_fraction=max_mask_fraction,
            p=1.0,
        )
        augmented_spectrogram = transform(magnitude_spectrogram)

        if DEBUG:
            plot_matrix(np.log(augmented_spectrogram))

        num_masked_frequencies = 0
        for i in range(augmented_spectrogram.shape[0]):
            frequency_slice = augmented_spectrogram[i]
            if (
                np.amin(frequency_slice) == np.amax(frequency_slice)
                and sum(frequency_slice) != 0.0
            ):
                num_masked_frequencies += 1

        assert num_masked_frequencies >= int(
            round(magnitude_spectrogram.shape[0] * min_mask_fraction)
        )
        assert num_masked_frequencies <= int(
            round(magnitude_spectrogram.shape[0] * max_mask_fraction)
        )

    def test_fill_mean_multichannel(self):
        samples, sample_rate = load_sound_file(
            os.path.join(DEMO_DIR, "background_noises", "hens.ogg"),
            sample_rate=None,
            mono=False,
        )
        assert samples.shape[0] == 2
        magnitude_spectrogram_chn0 = librosa.feature.melspectrogram(
            y=np.asfortranarray(samples[0, :]), sr=sample_rate
        )
        magnitude_spectrogram_chn1 = librosa.feature.melspectrogram(
            y=np.asfortranarray(samples[1, :]), sr=sample_rate
        )
        multichannel_magnitude_spectrogram = np.zeros(
            shape=(
                magnitude_spectrogram_chn0.shape[0],
                magnitude_spectrogram_chn0.shape[1],
                3,
            ),
            dtype=np.float32,
        )
        multichannel_magnitude_spectrogram[:, :, 0] = magnitude_spectrogram_chn0
        multichannel_magnitude_spectrogram[:, :, 1] = magnitude_spectrogram_chn1
        multichannel_magnitude_spectrogram[:, :, 2] = magnitude_spectrogram_chn1

        mask_fraction = 0.05
        transform = SpecFrequencyMask(
            fill_mode="mean",
            min_mask_fraction=mask_fraction,
            max_mask_fraction=mask_fraction,
            p=1.0,
        )
        augmented_spectrogram = transform(multichannel_magnitude_spectrogram)

        if DEBUG:
            image = (7 + np.log10(augmented_spectrogram + 0.0000001)) / 8
            plot_matrix(image)

        with np.testing.assert_raises(AssertionError):
            np.testing.assert_array_equal(
                augmented_spectrogram, multichannel_magnitude_spectrogram
            )

        num_masked_frequencies = 0
        for i in range(augmented_spectrogram.shape[0]):
            frequency_slice = augmented_spectrogram[i]
            if (
                np.amin(frequency_slice) == np.amax(frequency_slice)
                and np.sum(frequency_slice) != 0.0
            ):
                num_masked_frequencies += 1

        assert num_masked_frequencies == int(
            round(multichannel_magnitude_spectrogram.shape[0] * mask_fraction)
        )



import random

import numpy as np
import pytest

from audiomentations import GainTransition


class TestGainTransition:
    @pytest.mark.parametrize(
        "samples",
        [
            # Test both mono and stereo
            np.random.uniform(low=-0.5, high=0.5, size=(1234,)).astype(np.float32),
            np.random.uniform(low=-0.5, high=0.5, size=(2, 5678)).astype(np.float32),
        ],
    )
    def test_gain_transition_fraction(self, samples):
        np.random.seed(42)
        random.seed(42)
        sample_rate = 8000

        augment = GainTransition(p=1.0)
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert not np.allclose(samples, processed_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

    def test_gain_transition_seconds(self):
        np.random.seed(40)
        random.seed(40)
        samples = np.random.uniform(low=-0.5, high=0.5, size=(2345,)).astype(np.float32)
        sample_rate = 16000

        augment = GainTransition(
            min_duration=0.2, max_duration=0.3, duration_unit="seconds", p=1.0
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert not np.allclose(samples, processed_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

    def test_gain_transition_samples(self):
        np.random.seed(1337)
        random.seed(1337)
        samples = np.random.uniform(low=-0.5, high=0.5, size=(3456,)).astype(np.float32)
        sample_rate = 32000

        augment = GainTransition(
            min_duration=1, max_duration=5000, duration_unit="samples", p=1.0
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert not np.allclose(samples, processed_samples)
        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

    def test_invalid_params(self):
        with pytest.raises(ValueError):
            GainTransition(
                min_duration=-12, max_duration=324, duration_unit="samples", p=1.0
            )

        with pytest.raises(ValueError):
            GainTransition(
                min_duration=45, max_duration=10, duration_unit="samples", p=1.0
            )

        with pytest.raises(ValueError):
            GainTransition(
                min_gain_db=6.0, max_gain_db=-24.0, p=1.0
            )

        augment = GainTransition(
            min_duration=45, max_duration=45, duration_unit="lightyears", p=1.0
        )
        with pytest.raises(ValueError):
            augment(np.zeros(40, dtype=np.float32), 16000)



import random
import warnings
from typing import Optional, Callable

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import get_crossfade_mask_pair

# 0.00025 seconds corresponds to 2 samples at 8000 Hz
DURATION_EPSILON = 0.00025


class RepeatPart(BaseWaveformTransform):
    """
    Select a subsection (or "part") of the audio and repeat that part a number of times.
    This can be useful when simulating scenarios where a short audio snippet gets
    repeated, for example:

    * Repetitions of some musical note or sound in a rhythmical way
    * A person stutters or says the same word (with variations) multiple times in a row
    * A mechanical noise with periodic repetitions
    * A "skip in the record" or a "stuck needle" effect, reminiscent of vinyl records or
        CDs when they repeatedly play a short section due to a scratch or other
        imperfection.
    * Digital audio glitches, such as a buffer underrun in video games,
        where the current audio frame gets looped continuously due to system overloads
        or a software crash.

    Note that the length of inputs you give it must be compatible with the part
    duration range and crossfade duration. If you give it an input audio array that is
    too short, a `UserWarning` will be raised and no operation is applied to the signal.
    """

    supports_multichannel = True

    def __init__(
        self,
        min_repeats: int = 1,
        max_repeats: int = 3,
        min_part_duration: float = 0.25,
        max_part_duration: float = 1.2,
        mode: str = "insert",
        crossfade_duration: float = 0.005,
        part_transform: Optional[
            Callable[[NDArray[np.float32], int], NDArray[np.float32]]
        ] = None,
        p: float = 0.5,
    ):
        """
        :param min_repeats: Minimum number of times a selected audio segment should be
            repeated in addition to the original. For instance, if the selected number
            of repeats is 1, the selected segment will be followed by one repeat.
        :param max_repeats: Maximum number of times a selected audio segment can be
            repeated in addition to the original.
        :param min_part_duration: Minimum duration (in seconds) of the audio segment
            that can be selected for repetition.
        :param max_part_duration: Maximum duration (in seconds) of the audio segment
            that can be selected for repetition.
        :param mode: This parameter has two options:
            "insert": Insert the repeat(s), making the array longer. After the last
                repeat there will be the last part of the original audio, offset in time
                compared to the input array.
            "replace": Have the repeats replace (as in overwrite) the original audio.
                Any remaining part at the end (if not overwritten by repeats) will be
                left untouched without offset. The length of the output array is the
                same as the input array.
        :param crossfade_duration: Duration (in seconds) for crossfading between repeated
            parts as well as potentially from the original audio to the repetitions and back.
            The crossfades will be equal-energy or equal-gain depending on the audio and/or the
            chosen parameters of the transform. The crossfading feature can be used to smooth
            transitions and avoid abrupt changes, which can lead to impulses/clicks in the audio.
            If you know what you're doing, and impulses/clicks are desired for your use case,
            you can disable the crossfading by setting this value to `0.0`.
        :param part_transform: An optional callable (audiomentations transform) that
            gets applied individually to each repeat. This can be used to make each
            repeat slightly different from the previous one. Note that a part_transform
            that makes the part shorter is only supported if the transformed part is at
            least two times the crossfade duration.
        :param p: The probability of applying this transform
        """
        super().__init__(p)

        if min_repeats < 1:
            raise ValueError("min_repeats must be >= 1")
        if max_repeats < min_repeats:
            raise ValueError("max_repeats must be >= min_repeats")
        self.min_repeats = min_repeats
        self.max_repeats = max_repeats
        if min_part_duration < DURATION_EPSILON:
            raise ValueError(f"min_part_duration must be >= {DURATION_EPSILON}")
        if max_part_duration < min_part_duration:
            raise ValueError("max_part_duration must be >= min_part_duration")
        self.min_part_duration = min_part_duration
        self.max_part_duration = max_part_duration
        if mode not in ("insert", "replace"):
            raise ValueError('mode must be set to either "insert" or "replace"')
        self.mode = mode

        if crossfade_duration == 0.0:
            self.crossfade = False
        elif crossfade_duration < 0.0:
            raise ValueError("crossfade_duration must not be negative")
        elif crossfade_duration < DURATION_EPSILON:
            raise ValueError(
                "When crossfade_duration is set to a positive number, it must be >="
                f" {DURATION_EPSILON}"
            )
        else:
            self.crossfade = True
        if crossfade_duration > (min_part_duration / 2):
            raise ValueError(
                "crossfade_duration must be <= 0.5 * min_part_duration. You can fix this"
                " error by increasing min_part_duration or by decreasing"
                " crossfade_duration."
            )
        self.crossfade_duration = crossfade_duration
        self.part_transform = part_transform

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["part_num_samples"] = random.randint(
                int(self.min_part_duration * sample_rate),
                int(self.max_part_duration * sample_rate),
            )

            crossfade_length = self.get_crossfade_length(sample_rate)
            half_crossfade_length = crossfade_length // 2
            if (
                half_crossfade_length
                >= samples.shape[-1] - self.parameters["part_num_samples"]
            ):
                warnings.warn(
                    "The input sound is not long enough for applying the RepeatPart"
                    " transform with the current parameters."
                )
                self.parameters["should_apply"] = False
                return

            self.parameters["part_start_index"] = random.randint(
                half_crossfade_length,
                samples.shape[-1] - self.parameters["part_num_samples"],
            )
            self.parameters["repeats"] = random.randint(
                self.min_repeats, self.max_repeats
            )

    def get_crossfade_length(self, sample_rate: int) -> int:
        if not self.crossfade:
            return 0
        crossfade_length = int(self.crossfade_duration * sample_rate)
        if crossfade_length < 2:
            warnings.warn(
                "crossfade_duration is too small for the given sample rate. Using a"
                " crossfade length of 2 samples."
            )
            crossfade_length = 2
        elif crossfade_length % 2 == 1:
            crossfade_length += 1
        return crossfade_length

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        crossfade_length = 0
        half_crossfade_length = 0
        equal_energy_fade_in_mask = None
        equal_energy_fade_out_mask = None
        last_crossfade_type = "equal_energy"
        if self.crossfade:
            crossfade_length = self.get_crossfade_length(sample_rate)
            half_crossfade_length = crossfade_length // 2
            equal_energy_fade_in_mask, equal_energy_fade_out_mask = (
                get_crossfade_mask_pair(crossfade_length)
            )

        if self.crossfade:
            part = samples[
                ...,
                self.parameters["part_start_index"]
                - half_crossfade_length : self.parameters["part_start_index"]
                + self.parameters["part_num_samples"]
                + half_crossfade_length,
            ]
        else:
            part = samples[
                ...,
                self.parameters["part_start_index"] : self.parameters[
                    "part_start_index"
                ]
                + self.parameters["part_num_samples"],
            ]

        repeats_start_index = (
            self.parameters["part_start_index"] + self.parameters["part_num_samples"]
        )

        last_repetition_end_index = repeats_start_index
        parts = []
        for i in range(self.parameters["repeats"]):
            start_idx = last_repetition_end_index
            if self.crossfade:
                if i == 0:
                    start_idx -= half_crossfade_length
                else:
                    start_idx -= crossfade_length

            part_array = np.copy(part)

            if self.part_transform:
                part_array = self.part_transform(part_array, sample_rate)
                if self.crossfade and part_array.shape[-1] < 2 * crossfade_length:
                    raise ValueError(
                        "Applying a part_transform that makes a part shorter than 2 *"
                        " crossfade_duration is not supported"
                    )

            last_repetition_end_index = start_idx + part_array.shape[-1]

            if self.crossfade:
                part_array[..., :crossfade_length] *= equal_energy_fade_in_mask

                fade_out_crossfade_type = "equal_energy"
                is_last_part = i == self.parameters["repeats"] - 1
                if is_last_part and self.mode == "insert":
                    if self.part_transform is None:
                        fade_out_crossfade_type = "equal_gain"
                    else:
                        # If the part was transformed, check if it still correlates with
                        # the original signal in the seam. If it does, we use equal-gain
                        # crossfading instead of equal-energy.
                        correlation_coefficient = np.corrcoef(
                            part_array[..., -crossfade_length:],
                            samples[
                                ...,
                                self.parameters["part_start_index"]
                                + self.parameters["part_num_samples"]
                                - half_crossfade_length : self.parameters[
                                    "part_start_index"
                                ]
                                + self.parameters["part_num_samples"]
                                + half_crossfade_length,
                            ],
                        )[0, 1]
                        if abs(correlation_coefficient) > 0.5:
                            fade_out_crossfade_type = "equal_gain"
                if is_last_part:
                    last_crossfade_type = fade_out_crossfade_type
                if fade_out_crossfade_type == "equal_energy":
                    part_array[..., -crossfade_length:] *= equal_energy_fade_out_mask
                else:
                    _, equal_gain_fade_out_mask = get_crossfade_mask_pair(
                        crossfade_length, equal_energy=False
                    )
                    part_array[..., -crossfade_length:] *= equal_gain_fade_out_mask

            parts.append(
                {
                    "array": part_array,
                    "start_idx": start_idx,
                    "end_idx": last_repetition_end_index,
                }
            )
            if self.mode == "replace" and last_repetition_end_index > samples.shape[-1]:
                break

        result_length = samples.shape[-1]
        if self.mode == "insert":
            result_length += (
                parts[-1]["end_idx"] - parts[0]["start_idx"] - crossfade_length
            )

        if samples.ndim == 1:
            result_shape = (result_length,)
        else:
            result_shape = (samples.shape[0], result_length)

        result_placeholder = np.zeros(shape=result_shape, dtype=np.float32)

        if self.crossfade:
            # Fade out the signal where the first repetition takes over
            result_placeholder[..., : repeats_start_index - half_crossfade_length] = (
                samples[..., : repeats_start_index - half_crossfade_length]
            )
            result_placeholder[
                ...,
                repeats_start_index
                - half_crossfade_length : repeats_start_index
                + half_crossfade_length,
            ] = (
                equal_energy_fade_out_mask
                * samples[
                    ...,
                    repeats_start_index
                    - half_crossfade_length : repeats_start_index
                    + half_crossfade_length,
                ]
            )
        else:
            result_placeholder[..., :repeats_start_index] = samples[
                ..., :repeats_start_index
            ]

        # Add all repetitions except the last one
        for part in parts[:-1]:
            result_placeholder[..., part["start_idx"] : part["end_idx"]] += part[
                "array"
            ]

        # Add the (potentially truncated) last repetition
        if self.mode == "replace" and parts[-1]["end_idx"] > samples.shape[-1]:
            truncated_part_length = samples.shape[-1] - parts[-1]["start_idx"]
            truncated_part = parts[-1]["array"][..., :truncated_part_length]
            result_placeholder[
                ...,
                parts[-1]["start_idx"] : parts[-1]["start_idx"] + truncated_part_length,
            ] += truncated_part
        else:
            result_placeholder[
                ..., parts[-1]["start_idx"] : parts[-1]["end_idx"]
            ] += parts[-1]["array"]

        del parts

        if self.mode == "insert":
            if self.crossfade:
                if last_crossfade_type == "equal_energy":
                    fade_in_mask = equal_energy_fade_in_mask
                else:
                    fade_in_mask, _ = get_crossfade_mask_pair(
                        crossfade_length, equal_energy=False
                    )
                result_placeholder[
                    ...,
                    last_repetition_end_index
                    - crossfade_length : last_repetition_end_index,
                ] += (
                    fade_in_mask
                    * samples[
                        ...,
                        -(result_length - last_repetition_end_index)
                        - crossfade_length : -(
                            result_length - last_repetition_end_index
                        ),
                    ]
                )

            result_placeholder[..., last_repetition_end_index:] = samples[
                ..., -(result_length - last_repetition_end_index) :
            ]
        else:
            if self.crossfade:
                last_crossfade_start_idx = last_repetition_end_index - crossfade_length
                if last_crossfade_start_idx < samples.shape[-1]:
                    last_crossfade_length = crossfade_length
                    if last_repetition_end_index > samples.shape[-1]:
                        # Truncate crossfade
                        last_crossfade_length = (
                            samples.shape[-1] - last_crossfade_start_idx
                        )

                    result_placeholder[
                        ...,
                        last_crossfade_start_idx : last_crossfade_start_idx
                        + last_crossfade_length,
                    ] += (
                        equal_energy_fade_in_mask[:last_crossfade_length]
                        * samples[
                            ...,
                            last_crossfade_start_idx : last_crossfade_start_idx
                            + last_crossfade_length,
                        ]
                    )

            if last_repetition_end_index < result_length:
                result_placeholder[..., last_repetition_end_index:] = samples[
                    ..., last_repetition_end_index:
                ]
        return result_placeholder

    def freeze_parameters(self):
        super().freeze_parameters()
        if hasattr(self.part_transform, "freeze_parameters"):
            self.part_transform.freeze_parameters()

    def unfreeze_parameters(self):
        super().unfreeze_parameters()
        if hasattr(self.part_transform, "unfreeze_parameters"):
            self.part_transform.unfreeze_parameters()
